"""
fprime_gds.executables.utils:

Utility functions to enable the executables package to function seamlessly.
"""
import atexit
import signal
import subprocess
import time

# Python 2.7 compatibility, adding in missing error type
try:
    InterruptedError
except NameError:

    class InterruptedError(Exception):
        pass


class ProcessNotStableException(Exception):
    """Process did not start up stably. Thus there was a problem."""

    def __init__(self, name, code, lifespan):
        """Constructor to help with messages"""
        super().__init__(
            f"{name} stopped with code {code} sooner than {lifespan} seconds"
        )


class AppWrapperException(Exception):
    """
    An exception occurred while tying to start the app wrapper. This will encapsulate that message.
    """


def register_process_assassin(process, log=None):
    """
    Register an assassin that will kill the a given child process when an exit of the current python process has been
    reached. This will effectively clean up children and (optionally) their log files.

    :param process: the process to kill.
    :param log: a paired log file to kill as well.
    """

    def assassin():
        """
        Kill process and ensure that it is really really dead.
        Note: enclosing the locals in the closure, allows this function to operate independently.
        """
        # First attempt to kill the process uses SIGINT/SIGTERM giving the process a bit to wrap up its affairs.
        # This code allows for both pexpect and subprocess processes.
        try:
            if hasattr(process, "terminate"):
                process.terminate()
            else:
                process.kill(signal.SIGINT)
            time.sleep(1)
        except (KeyboardInterrupt, OSError, InterruptedError):
            pass
        # Second attempt is to terminate with extreme prejudice. No process will survive this, ensuring that it is
        # really, really dead. Supports both pexpect and subprocess.
        try:
            if hasattr(process, "terminate"):
                process.kill()
            else:
                process.kill(signal.SIGKILL)
        except (KeyboardInterrupt, OSError, InterruptedError):
            pass
        # Might as well close the log file because dead men tell no tales.
        try:
            if log is not None:
                log.close()
        except (KeyboardInterrupt, OSError, InterruptedError):
            pass

    atexit.register(assassin)


def run_wrapped_application(arguments, logfile=None, env=None, launch_time=None):
    """
    Run an application and ensure that it is logged immediately to the logfile. This will allow the application to have
    up-to-date logs. This is a wrapper for pexpect to ensure that the application runs and log effectively. It has been
    converted to a function to remove superfluous processes.

    :param arguments: arguments with the first being the executable.
    :param logfile: (optional) path to logfile to log to. Will overwrite.
    :param env: (optional) environment for the subprocess
    :param launch_time: (optional) time to wait before declaring the process stable
    :return: child process should it be needed.
    """
    # Write out run information for the calling user
    print(f"[INFO] Running Application: {arguments[0]}")
    # Attempt to open a log file
    file_handler = None
    try:
        if logfile is not None:
            print(f"[INFO] Log File: {logfile}")
            file_handler = open(logfile, "wb", 0)
    except OSError as exc:
        raise AppWrapperException(f"Failed to open: {logfile} with error {str(exc)}.")
    # Spawn the process. Uses pexpect, as this will force the process to output data immediately, rather than buffering
    # the output. That way the log file is fully up-to-date.
    try:
        child = subprocess.Popen(
            arguments, stdout=file_handler, stderr=subprocess.STDOUT, env=env
        )
        register_process_assassin(child, file_handler)
        # If launch time is specified, then wait for it to be stable
        if launch_time is not None:
            time.sleep(launch_time)
            child.poll()
            if child.returncode is not None:
                raise ProcessNotStableException(
                    arguments[0], child.returncode, launch_time
                )
        return child
    except Exception as exc:
        raise AppWrapperException(
            f'Failed to run application: {" ".join(arguments)}. Error: {exc}'
        )



#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK

"""
Parses the given CLI command for the F' GDS, and then executes the correct
command with the user-provided arguments on the GDS
"""

import abc
import argparse
from copy import deepcopy
import os
import sys
from typing import Callable, List, Union

import argcomplete

# NOTE: These modules are now only lazily loaded below as needed, due to slow
# performance when importing them
# import fprime_gds.common.gds_cli.channels as channels
# import fprime_gds.common.gds_cli.command_send as command_send
# import fprime_gds.common.gds_cli.events as events
# from fprime_gds.common.pipeline.dictionaries import Dictionaries
from fprime_gds.executables.cli import GdsParser


def add_connection_arguments(parser: argparse.ArgumentParser):
    """
    Adds the arguments needed to properly connect to the API, which the user may
    want to specify
    """
    parser.add_argument(
        "-d",
        "--dictionary",
        type=str,
        default=None,
        help='path from the current working directory to the "<project name>Dictionary.xml" file for the project you\'re using the API with; if unused, tries to search the current working directory for such a file',
    )
    parser.add_argument(
        "-ip",
        "--ip-address",
        type=str,
        default="127.0.0.1",
        help="connect to the GDS server using the given IP or hostname (default=127.0.0.1 (i.e. localhost))",
        metavar="IP",
    )
    parser.add_argument(
        "-p",
        "--port",
        type=int,
        default=50050,
        help="connect to the GDS server using the given port number (default=50050)",
        metavar="PORT",
    )


def add_retrieval_arguments(parser: argparse.ArgumentParser):
    """
    Adds all the arguments that affect how a commands retrieves a message generated by the GDS or F' instance
    """
    parser.add_argument(
        "-t",
        "--timeout",
        type=float,
        default=0.0,
        help="wait at most SECONDS seconds for a single new message, then exit (defaults to listening until the user exits via CTRL+C, and logging all messages)",
        metavar="SECONDS",
    )


def add_search_arguments(parser: argparse.ArgumentParser, command_name: str):
    """
    Adds all the arguments relevant to searching/filtering certain messages
    used by the Channels/Commands/Events commands with the given name used in
    the help text, due to the similarity of each of these commands
    """
    parser.add_argument(
        "-l",
        "--list",
        dest="is_printing_list",
        action="store_true",
        help=f"list all possible {command_name[:-1]} types the current F Prime instance could produce, based on the {command_name} dictionary, sorted by {command_name[:-1]} type ID",
    )
    parser.add_argument(
        "-i",
        "--ids",
        nargs="+",
        type=int,
        help=f'only show {command_name} matching the given type ID(s) "ID"; can provide multiple IDs to show all given types',
        metavar="ID",
    )
    parser.add_argument(
        "-c",
        "--components",
        nargs="+",
        type=str,
        help=f'only show {command_name} from the given component name "COMP"; can provide multiple components to show {command_name} from all components given',
        metavar="COMP",
    )
    parser.add_argument(
        "-s",
        "--search",
        type=str,
        help=f'only show {command_name} whose name or output string exactly matches or contains the entire given string "STRING"',
        metavar="STRING",
    )
    parser.add_argument(
        "-j",
        "--json",
        action="store_true",
        help=f"return the JSON response of the API call, with {command_name} filtered based on other flags provided",
    )


def get_dictionary_path(current_args: argparse.Namespace) -> Union[str, None]:
    """
    Returns the current project dictionary, either one provided by the user
    or the first one found in the current working directory. Raises an
    exception if neither one is found.
    """
    args = deepcopy(current_args)
    if not hasattr(args, "dictionary"):
        args.dictionary = None
    args.deploy = os.getcwd()
    args.config = None
    args = GdsParser.handle_arguments(args, kwargs={})
    return args.dictionary


def add_valid_dictionary(args: argparse.Namespace) -> argparse.Namespace:
    """
    Takes in the given parsed arguments and, if no F' dictionary has been given,
    attempt to search for one in the current working directory. Throw an error
    if none can be found OR if the given dictionary is invalid.
    """
    args.dictionary = get_dictionary_path(args)
    return args


class CliCommandParserBase(abc.ABC):
    """
    An abstract class for CLI commands to implement; provides methods for
    adding a new parser for this command
    """

    @classmethod
    def add_subparser(cls, parent_parser: argparse.ArgumentParser):
        """
        Adds this command as a sub-command to an existing parser, so that it
        can be passed in as an argument (similar to git's CLI tool)
        """
        command_parser = cls.create_subparser(parent_parser)
        cls.add_arguments(command_parser)
        command_parser.set_defaults(func=cls.command_func, validate=cls.validate_args)

    @classmethod
    @abc.abstractmethod
    def create_subparser(
        cls, parent_parser: argparse.ArgumentParser
    ) -> argparse.ArgumentParser:
        """
        Creates the parser for this command as a subparser of the given one,
        and then returns it
        """

    @classmethod
    @abc.abstractmethod
    def add_arguments(cls, parser: argparse.ArgumentParser):
        """
        Add all the required and optional arguments for this command to the
        given parser
        """

    @classmethod
    def validate_args(cls, parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Validates the parsed arguments for this parser; if any are incorrect or
        missing, try to set them automatically if possible and error if this is
        not possible

        By default, tries to set the project dictionary
        """
        try:
            args = add_valid_dictionary(args)
        except ValueError:
            parser.error("No valid project dictionary found")
        return args

    @classmethod
    @abc.abstractmethod
    def command_func(cls, *args, **kwargs) -> Callable:
        """
        Executes the appropriate function when this command is called
        """


class ChannelsParser(CliCommandParserBase):
    """
    A parser for the "channels" CLI command, which lets users retrieve
    information about recent telemetry data from an F' instance
    """

    @classmethod
    def create_subparser(cls, parent_parser: argparse.ArgumentParser):
        """
        Creates the channels sub-command as a subparser, and then returns it
        """
        channels_parser = parent_parser.add_parser(
            "channels",
            description="print out new telemetry data that has been received from the F Prime instance, sorted by timestamp",
        )
        return channels_parser

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser):
        """
        Add all the required and optional arguments for this command to the
        given parser
        """
        add_connection_arguments(parser)
        add_retrieval_arguments(parser)
        add_search_arguments(parser, "channels")

    @classmethod
    def command_func(cls, *args, **kwargs):
        """
        Executes the appropriate function when "channels" is called
        """
        import fprime_gds.common.gds_cli.channels as channels

        channels.ChannelsCommand.handle_arguments(*args, **kwargs)


class CommandSendParser(CliCommandParserBase):
    """
    A parser for the "command-send" CLI command, which lets users send commands
    from the GDS to a running F' instance and lets retrieve information about
    what commands are available
    """

    @classmethod
    def create_subparser(cls, parent_parser: argparse.ArgumentParser):
        """
        Creates the command-send sub-command as a subparser, and then returns it
        """
        command_send_parser = parent_parser.add_parser(
            "command-send",
            description="sends the given command to the spacecraft via the GDS",
        )
        return command_send_parser

    @staticmethod
    def complete_command_name(
        prefix: str, parsed_args: argparse.Namespace, **kwargs
    ) -> List[str]:
        """
        Returns a list of all command names that could possibly complete the
        given prefix
        """
        # Kwargs arguments required for argcomplete, so suppress warning
        # pylint: disable=unused-argument
        dict_path = ""
        try:
            dict_path = get_dictionary_path(parsed_args)
        except ValueError:
            argcomplete.warn("No dictionary found to get command names from")
            return []

        from fprime_gds.common.pipeline.dictionaries import Dictionaries

        dictionary = Dictionaries()
        dictionary.load_dictionaries(dict_path, None)
        command_names = dictionary.command_name.keys()
        return [name for name in command_names if name.startswith(prefix)]

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser):
        """
        Add all the required and optional arguments for this command to the
        given parser
        """
        add_connection_arguments(parser)
        parser.add_argument(
            "command_name",
            help='the full name of the command you want to execute in "<component>.<name>" form',
            nargs="?",
            metavar="command-name",
        ).completer = cls.complete_command_name
        # NOTE: Type set to string because we don't know the type beforehand
        parser.add_argument(
            "-args",
            "--arguments",
            nargs="*",
            type=str,
            default=[],
            help="provide a space-separated set of arguments to the command being sent",
        )
        add_search_arguments(parser, "commands")

    @classmethod
    def validate_args(cls, parser: argparse.ArgumentParser, args: argparse.Namespace):
        """
        Validates the parsed arguments for command_send; if any are incorrect or
        missing, try to set them automatically if possible and error if this is
        not possible
        """
        if not (args.command_name or args.is_printing_list):
            parser.error("One of command-name or -l/--list is required")

        args = super().validate_args(parser, args)
        return args

    @classmethod
    def command_func(cls, *args, **kwargs):
        """
        Executes the appropriate function when "command_send" is called
        """
        import fprime_gds.common.gds_cli.command_send as command_send

        command_send.CommandSendCommand.handle_arguments(*args, **kwargs)


class EventsParser(CliCommandParserBase):
    """
    A parser for the "events" CLI command, which lets users retrieve
    information about recent events logged on an F' instance
    """

    @classmethod
    def create_subparser(cls, parent_parser: argparse.ArgumentParser):
        """
        Creates the events sub-command as a subparser, and then returns it
        """
        events_parser = parent_parser.add_parser(
            "events",
            description="print out new events that have occurred on the F Prime instance, sorted by timestamp",
        )
        return events_parser

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser):
        """
        Add all the required and optional arguments for this command to the
        given parser
        """
        add_connection_arguments(parser)
        add_retrieval_arguments(parser)
        add_search_arguments(parser, "events")

    @classmethod
    def command_func(cls, *args, **kwargs):
        """
        Executes the appropriate function when "events_parser" is called
        """
        import fprime_gds.common.gds_cli.events as events

        events.EventsCommand.handle_arguments(*args, **kwargs)


def create_parser():
    parser = argparse.ArgumentParser(
        description="provides utilities for interacting with the F' Ground Data System (GDS)"
    )
    parser.add_argument("-V", "--version", action="version", version="0.0.2")

    # Add subcommands to the parser
    subparsers = parser.add_subparsers(dest="func")

    ChannelsParser.add_subparser(subparsers)
    CommandSendParser.add_subparser(subparsers)
    EventsParser.add_subparser(subparsers)

    return parser


def parse_args(parser: argparse.ArgumentParser, arguments):
    """
    Parses the given arguments and returns the resulting namespace; having this
    separate allows for unit testing if needed
    """
    args = parser.parse_args(arguments)

    if not args.func:
        # no argument provided, so print the help message
        parser.print_help()
        sys.exit()

    args = args.validate(parser, args)

    return args


def main():
    # parse arguments, not including the name of this script
    parser = create_parser()
    argcomplete.autocomplete(parser)
    args = parse_args(parser, sys.argv[1:])

    # call the selected function with the args provided
    function = args.func
    argument_dict = vars(args)
    # Remove function and argument validation args
    del argument_dict["func"]
    del argument_dict["validate"]
    function(**argument_dict)


if __name__ == "__main__":
    main()



"""
fprime_gds.common.transport:

Sets up the classes used to transport data within the GDS system between the few executables that compose the GDS. This
file defines several base classes used to specify what a client looks like and it defines a basic implementation:
ThreadedTCPSocketClient used to send data through the threaded tcp server.

@author lestarch
"""
from abc import abstractmethod, ABC
from enum import Enum
import select
import socket
import threading

from fprime_gds.common.handlers import DataHandler, HandlerRegistrar


class RoutingTag(Enum):
    """Tag for routing data about the system"""

    GUI = b"GUI"
    FSW = b"FSW"


class TransportationException(Exception):
    pass


class TransportClient(DataHandler, HandlerRegistrar, ABC):
    """Transport client used as an interface for handling transportation within the GDS

    The GDS is composed of multiple parts: communications talking to the embedded software, the GUI displaying
    information to a user, etc. Between these various executables and functions there needs to be data transportation.
    This class acts as an interface for various implementations that shuttle that data about the GDS.

    This has two important parents: DataHandler and HandlerRegistrar. The DataHandler allows this to be registered to
    anything that produces data making sending data out of this client east. The HandlerRegistrar allows this to pass
    incoming data to various handlers easy.
    """

    @abstractmethod
    def connect(
        self,
        connection_uri,
        incoming_routing=RoutingTag.GUI,
        outgoing_routing=RoutingTag.FSW,
    ):
        """Connect to this client given the connection URI

        For clients that can handle a URI directly this should be consumed as-is. For clients expecting a host/port this
        should be split on ':' to break apart the URI. The connection needs to specify routing (e.g. FSW, GUI) for both
        incoming data and outgoing data. Since this is a client it defaults to the incoming GUI tag, and outgoing FSW
        tag.  It is up to the client to use this information.

        Args:
            connection_uri: connection URI for setting up this client
            incoming_routing: routing tag for the incoming (received) data
            outgoing_routing: routing tag for the outgoing (sent) data
        """

    @abstractmethod
    def disconnect(self):
        """Disconnects from this client ensuring all resources deallocated"""

    @abstractmethod
    def send(self, data):
        """Send the supplied data out of this client

        Sends the supplied data out of this client. This data is not specifically formatted. It is expected that the
        data is a set of bytes to send. This is send method is typically triggered by the data_callback method and will
        send any data received by that call. Send is not governed by a timeout and will block until the send succeeds.

        Args:
            data {bytes}: data to send out this client
        """

    @abstractmethod
    def recv(self, timeout=None):
        """Receive data from the interface and return it

        Receives the data from the client and returns it out of the interface. The client implementation is not expected
        to return complete messages. It may or may not do so and the calling function should expect misc bytes returned.
        Recv is governed by a timeout such that it does not hand forever.

        Args:
            timeout {int}: timeout in milliseconds before returning b""
        Returns:
             byte data read from interface. b"" on timeout.
        """

    def data_callback(self, data, sender=None):
        """Calls send to send received data"""
        assert isinstance(
            data, (bytes, bytearray)
        ), "Client cannot handle non-binary data callbacks"
        self.send(data)


class ThreadedTransportClient(TransportClient, ABC):
    """Transportation client that uses a thread to receive and dispatch calls through registrants

    This client will setup a thread and use this thread to call recv to receive data and then forward that data to all
    registrants. The thread is created at connection time and will stop after disconnect is called. Subclasses should
    call connect here after the connection is setup and should call disconnect (typically) before the connection is
    destroyed although it is understood that the threads will join before disconnect is complete.
    """

    def __init__(self, timeout=100):
        """Sets up the stop event and thread object

        Args:
            timeout: read timeout supplied to the recv call within the thread. Sets shutdown time.
        """
        super().__init__()
        self.__data_recv_thread = threading.Thread(target=self.recv_thread, name="TTSReceiverThread")
        self.__stop_event = threading.Event()
        self.timeout = timeout
        self.started = False

    def connect(self, *_, **__):
        """Starts up the recv thread when connected

        Upon connection this client starts up the receive thread. Subclasses should call super().connect(...) after the
        connection has been created as it will immediately start reading data. All arguments are ignored.
        """
        super().connect(*_, **__)
        if not self.__stop_event.is_set():
            self.started = True
            self.__data_recv_thread.start()

    def disconnect(self):
        """Stops and joins to thee recv thread when disconnected

        Upon disconnection this client will stop the thread and then join to it. The maximum time to join should be
        ~self.timeout. Subclasses should call super().disconnect(...) before the connection is closed as the thread will
        read data until it is stopped.
        """
        super().disconnect()
        self.__stop_event.set()
        if self.started:
            self.__data_recv_thread.join()

    def stop(self):
        """ Stop the receive thread without waiting """
        self.__stop_event.set()

    def poll(self):
        """ Poll the receive and process the result

        Poll the receive and process the results. This will be called by the receive thread, but is separated out for
        flexibility in case users wish to reduce the number of empty threads in the system.
        """
        data = self.recv(self.timeout)
        if len(data) > 0:
            self.send_to_all(data)

    def recv_thread(self):
        """Loops reading data and dispatching it to registrants

        Reads data while supplying timeout on a thread. After each read the thread is checked to shutdown and as such
        the timeout is roughly the maximum time the thread may take to shutdown after the shutdown event is issued. The
        data is only forwarded to registrants if the data is not empty as that indicates a timeout.
        """
        while not self.__stop_event.is_set():
            self.poll()


class ThreadedTCPSocketClient(ThreadedTransportClient):
    """
    Threaded TCP client that connects to the socket server that serves packets from the flight software. The threaded
    tcp server acts as the original middleware layer for transporting data. This client is designed specifically to
    source incoming data from that server and send outgoing data to that server.
    """

    def __init__(self, sock=None):
        """
        Threaded client socket constructor

        Keyword Arguments:
            sock {Socket} -- A socket for the client to use. Created own if None (default: {None})
        """
        super().__init__()
        self.dest = None
        self.chunk_size = 1024
        if sock is None:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        else:
            self.sock = sock

    def connect(
        self,
        connection_uri,
        incoming_routing=RoutingTag.GUI,
        outgoing_routing=RoutingTag.FSW,
    ):
        """Connect to host at given port and start the threaded recv method

        Connects to a running ThreadedTcpServer at the given connection_uri expected in the format tcp://host:port or
        simply host: port. This will also register with the ThreadedTcpServer using the incoming

        Arguments:
            connection_uri {string} -- connection uri of the form (tcp://)?host:port
            incoming_routing {RoutingTag}: routing tag for incoming data, used to register client to server
            outgoing_routing {RoutingTag}: routing tog applied to outgoing data when sent
        """
        try:
            self.dest = outgoing_routing.value
            connection_uri = connection_uri.split("//", 1)[-1]  # Remove leading tcp
            host, port = connection_uri.split(":", 1)
            port = int(port)

            self.sock.connect((host, port))
            self.sock.send(b"Register %s\n" % incoming_routing.value)
            super().connect(connection_uri, incoming_routing, outgoing_routing)
        except ValueError as vle:
            raise TransportationException(
                f"Failed to parse connection uri: {connection_uri}. {vle}"
            )
        except Exception as exc:
            raise TransportationException(
                f"Failed to connect to transportation layer at: {connection_uri}. {exc}"
            )

    def disconnect(self):
        """Disconnect the socket client"""
        super().disconnect()
        self.sock.close()

    def send(self, data):
        """Send data to the server

        Sends data out to the destination via the threaded tcp server. All necessary headers are added in this function
        such that the sever can process this send.

        Arguments:
            data {binary} -- the data to send (What you want the destination to receive)
        """
        assert self.dest is not None, "Cannot send data before connect call"
        self.sock.send(b"A5A5 %s %s" % (self.dest, data))

    def recv(self, timeout=100):
        """Receives data from the threaded tcp server

        Receives raw data from the threaded tcp server. This data is expected to have no headers and will be passed as
        received or in chunks back to the caller.

        Args:
            timeout {int}: timeout to wait for the socket to have data before returning b"". Default: 100ms.
        """
        assert self.dest is not None, "Cannot recv data before connect call"
        ready = select.select([self.sock], [], [], timeout)
        return self.sock.recv(self.chunk_size) if ready[0] else b""



"""
Tests format_string util

@Created on March 18, 2021
@janamian
"""

import unittest
from fprime_gds.common.utils.string_util import format_string_template


class TestFormatString(unittest.TestCase):
    def test_format_with_no_issue(self):
        template = "Opcode 0x%04X dispatched to port %d and value %f"
        values = (181, 8, 1.234)
        expected = "Opcode 0x00B5 dispatched to port 8 and value 1.234000"
        actual = format_string_template(template, values)
        self.assertEqual(expected, actual)

    def test_format_value_with_string_input_as_other_types(self):
        template = "Opcode 0x%04X dispatched to port %u and value %.2f"
        values = (181, "8", 1.234)
        expected = "Opcode 0x00B5 dispatched to port 8 and value 1.23"
        actual = format_string_template(template, values)
        self.assertEqual(expected, actual)

    def test_format_with_format_spec(self):
        template = "Opcode 0x%04X dispatched to port %04d and value %0.02f"
        values = (181, 8, 1.234)
        expected = "Opcode 0x00B5 dispatched to port 0008 and value 1.23"
        actual = format_string_template(template, values)
        self.assertEqual(expected, actual)

    def test_format_bad_case(self):
        template = "Opcode 0x%04X dispatched to port %04d and value %0.02f"
        values = ("181", "8", "0.123")
        with self.assertRaises(ValueError):
            self.assertEqual(format_string_template(template, values))

    def test_format_decimal_with_width_flag(self):
        template = "Decimals: %d %ld"
        values = (1977, 650000)
        expected = "Decimals: 1977 650000"
        actual = format_string_template(template, values)
        self.assertEqual(expected, actual)

    def test_format_preceding_with_blanks(self):
        template = "Preceding with blanks: %10d"
        values = (1977,)
        expected = "Preceding with blanks:       1977"
        actual = format_string_template(template, values)
        self.assertEqual(expected, actual)

    def test_format_preceding_with_zeros(self):
        template = "Preceding with zeros: %010d"
        values = (1977,)
        expected = "Preceding with zeros: 0000001977"
        actual = format_string_template(template, values)
        self.assertEqual(expected, actual)

    def test_format_some_different_radices(self):
        template = "Some different radices: %d %x %o %#x %#o"
        values = (100, 100, 100, 100, 100)
        # The alternate form causes a leading octal specifier ('0o') to be inserted before the first digit. This is different than C behavior
        # `See https://docs.python.org/3/library/stdtypes.html#printf-style-bytes-formatting`
        expected = "Some different radices: 100 64 144 0x64 0o144"
        actual = format_string_template(template, values)
        self.assertEqual(expected, actual)

    def test_format_floats(self):
        template = "floats: %4.2f %+.0e %E"
        values = (3.1416, 3.1416, 3.1416)
        expected = "floats: 3.14 +3e+00 3.141600E+00"
        actual = format_string_template(template, values)
        self.assertEqual(expected, actual)

    def test_format_asterisk_width(self):
        # asterisk `*` is not supported by python string-format
        template = "Width trick: %*d"
        values = (5, 10)
        with self.assertRaises(ValueError):
            actual = format_string_template(template, values)

    def test_format_regular_string(self):
        template = "%s"
        values = ("A string",)
        expected = "A string"
        actual = format_string_template(template, values)
        self.assertEqual(expected, actual)

    def test_format_percent_sign(self):
        template = "%.2f%%"
        values = (1.23456,)
        expected = "1.23%"
        actual = format_string_template(template, values)
        self.assertEqual(expected, actual)

    def test_format_single_value(self):
        template = "%.2f%%"
        values = 1.23456
        expected = "1.23%"
        actual = format_string_template(template, values)
        self.assertEqual(expected, actual)

    def test_format_list_value(self):
        template = "%.2f%%, %.2f%%"
        values = [1.23456, 1.23456]
        expected = "1.23%, 1.23%"
        actual = format_string_template(template, values)
        self.assertEqual(expected, actual)

    def test_format_tuple_value(self):
        template = "%.2f%%, %.2f%%"
        values = (1.23456, 1.23456)
        expected = "1.23%, 1.23%"
        actual = format_string_template(template, values)
        self.assertEqual(expected, actual)

    def test_format_unsigned_flag_with_length_flag(self):
        template = "Something %lu something %llu something else %lu"
        values = (123456, 123457, 123458)
        expected = "Something 123456 something 123457 something else 123458"
        actual = format_string_template(template, values)
        self.assertEqual(expected, actual)


if __name__ == "__main__":
    unittest.main()



"""
ram.py:

A simple implementation of a history that maintains items in RAM. This is used for simplicity, but isn't exactly
robust nor persistent. Given that it is in the RAM, it is driven from the decoders object, which should run off the
middle-ware layer.

Note: this RAM history treats "start times" as session tokens to remember where it was last fetched from.

:author: lestarch
"""
import time
import threading

from fprime_gds.common.history.history import History


class RamHistory(History):
    """
    Chronological variant of history.  This is intended to be registered with the decoders in order
    to handle incoming objects, and store them for retrieval.
    """

    def __init__(self):
        """
        Constructor used to set-up in-memory store for history
        """
        self.lock = threading.RLock()
        self.objects = []
        self.retrieved_cursors = {}

    def data_callback(self, data, sender=None):
        """
        Data callback to store

        :param data: object to store
        """
        with self.lock:
            self.objects.append(data)

    def retrieve(self, start=None, limit=None):
        """
        Retrieve objects from this history. 'start' is the session token for retrieving new elements. If session is not
        specified, all elements are retrieved. If session is specified, then unseen elements are returned. If the
        session itself is new, it is recorded and set to the newest data.

        :param start: return all objects newer than given start session key
        :param limit: limit (count) of returned results
        :return: a list of objects
        """
        index = 0
        with self.lock:
            size = self.size()
            if start is not None:
                index = self.retrieved_cursors.get(start, size)
            end_slice = min(size, index + (limit if limit is not None else size))
            objs = self.objects[index:end_slice]
            self.retrieved_cursors[start] = end_slice
        return objs

    def retrieve_new(self):
        """
        Retrieves a chronological order of objects that haven't been accessed through retrieve or
        retrieve_new before.

        Returns:
            a list of objects in chronological order
        """
        index = 0
        with self.lock:
            if len(self.retrieved_cursors.values()) > 0:
                index = max(self.retrieved_cursors.values())
            return self.objects[index:]

    def clear(self, start=None):
        """
        Clears objects from RamHistory. It clears upto the earliest session. If session is supplied, the session id will
        be deleted as well.

        Args:
            start: a position in the history's order (int).
        """
        with self.lock:
            try:
                if start is not None:
                    del self.retrieved_cursors[start]
            except KeyError:
                pass
            earliest = 0
            if len(self.retrieved_cursors.values()) > 0:
                earliest = min(self.retrieved_cursors.values())
            del self.objects[:earliest]
            for key in self.retrieved_cursors.keys():
                self.retrieved_cursors[key] -= earliest

    def sessions(self):
        """
        Accessor for the number of stored sessions

        Returns:
            number of tracked sessions
        """
        return len(self.retrieved_cursors.values())

    def size(self):
        """
        Accessor for the number of objects in the history
        Returns:
            the number of objects (int)
        """
        with self.lock:
            return len(self.objects)


class SelfCleaningRamHistory(RamHistory):
    """A Ram history which clears itself after a time of inactivity"""

    def __init__(self):
        """Construct object"""
        super().__init__()
        self.last_request = {}
        self.clear_time = -1

    def set_clear_time(self, time):
        """Update the clear time"""
        self.clear_time = time

    def retrieve(self, start=None, limit=None):
        """
        Retrieve objects from this history. 'start' is the session token for retrieving new elements. If session is not
        specified, all elements are retrieved. If session is specified, then unseen elements are returned. If the
        session itself is new, it is recorded and set to the newest data. This refreshes the last polled time preventing
        self clearing for another time

        :param start: return all objects newer than given start session key
        :param limit: limit (count) of returned results
        :return: a list of objects
        """
        if start is not None:
            with self.lock:
                self.last_request[start] = time.time()
        return super().retrieve(start, limit)

    def clear(self, start=None):
        """
        Clears objects from RamHistory. It clears upto the earliest session. If session is supplied, the session id will
        be deleted as well. This will also check all sessions for expiration and clear any sessions that have not been
        updated in self.clear_time seconds, unless that value is negative.

        Args:
            start: a position in the history's order (int).
        """
        current = time.time()
        with self.lock:
            deletes = [
                key
                for key, last in self.last_request.items()
                if self.clear_time > 0 and (last + self.clear_time) < current
            ]
            for delete in deletes:
                for container in [self.retrieved_cursors, self.last_request]:
                    try:
                        del container[delete]
                    except KeyError:
                        pass
        return super().clear(start)



"""
@brief Encoder for Event data

This encoder takes in EventData objects, serializes them, and sends the results
to all registered senders.

Serialized Event format:
    +--------------------------------+          -
    | Header = "A5A5 "               |          |
    | (5 byte string)                |          |
    +--------------------------------+      Added by
    | Destination = "GUI " or "FSW " |       Sender
    | (4 byte string)                |          |
    +--------------------------------+          -
    | Length of descriptor, ID,      |
    | and argument data              |
    | (variable bytes, check config) |
    +--------------------------------+
    | Descriptor type = 2            |
    | (4 bytes)                      |
    +--------------------------------+
    | ID                             |
    | (4 bytes)                      |
    +--------------------------------+
    | Argument 1 value               |
    +--------------------------------+
    | Argument 2 value               |
    +--------------------------------+
    | ...                            |
    +--------------------------------+
    | Argument n value               |
    +--------------------------------+

@date Created August 9, 2018
@author R. Joseph Paetz

@bug No known bugs
"""

from .encoder import Encoder
from fprime_gds.common.data_types.event_data import EventData
from fprime_gds.common.utils.data_desc_type import DataDescType


class EventEncoder(Encoder):
    """Encoder class for event data"""

    def __init__(self, config=None):
        """
        Constructor

        Args:
            config (ConfigManager, default=None): Object with configuration data
                    for the sizes of fields in the binary data. If None passed,
                    defaults are used.

        Returns:
            An initialized EventEncoder object
        """
        super().__init__(config)

        self.len_obj = self.config.get_type("msg_len")
        self.desc_obj = self.config.get_type("msg_desc")
        self.id_obj = self.config.get_type("event_id")

    def encode_api(self, data):
        """
        Encodes the given EventData object as binary data and returns the result.

        Args:
            data (EventData obj): A non-empty EventData object to encode

        Returns:
            Encoded version of the data argument as binary data
        """
        assert isinstance(data, EventData), "Encoder handling incorrect type"
        event_temp = data.get_template()

        self.desc_obj.val = DataDescType["FW_PACKET_LOG"].value
        desc_bin = self.desc_obj.serialize()

        self.id_obj.val = event_temp.get_id()
        id_bin = self.id_obj.serialize()

        time_bin = data.get_time().serialize()

        arg_bin = b""
        for arg in data.get_args():
            arg_bin += arg.serialize()

        len_val = len(desc_bin) + len(id_bin) + len(time_bin) + len(arg_bin)
        self.len_obj.val = len_val
        len_bin = self.len_obj.serialize()

        binary_data = len_bin + desc_bin + id_bin + time_bin + arg_bin

        return binary_data



"""Configuration file for the Sphinx documentation builder.

Documentation:
https://www.sphinx-doc.org/en/master/usage/configuration.html
"""
# pylint: disable=invalid-name

import re
from datetime import datetime
from importlib import import_module
from pathlib import Path

import sphinx_rtd_theme  # pylint: disable=unused-import


# importlib.metadata is implemented in Python 3.8
# Previous versions require the backport, https://pypi.org/project/importlib-metadata/
try:
    metadata = import_module("importlib.metadata")
except ModuleNotFoundError:
    metadata = import_module("importlib_metadata")

# -- Project information -----------------------------------------------------

metadata_ = metadata.metadata("fprime-gds")
project = metadata_.get("Name")
author = metadata_.get("Author")
copyright = f"{datetime.now().year}, {author}"  # pylint: disable=redefined-builtin

# The full version, including alpha/beta/rc tags
release = metadata_.get("Version")
# The short X.Y version
version = re.match(r"v?\d+(\.\d+)*", release)[0]


# -- General configuration ---------------------------------------------------

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.doctest",
    "sphinx.ext.coverage",
    "sphinx.ext.extlinks",
    "sphinx.ext.ifconfig",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "autoapi.extension",
    "recommonmark",
    "sphinx_rtd_theme",
    "sphinxcontrib.mermaid",
]

source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}

master_doc = "index"

rst_epilog = "\n".join(
    [
        "\nBuild: |release|\n",
        f".. _fprime-gds: {metadata_.get('url')}",
        f".. |project| replace:: {project.replace('_', ' ').title()}",
    ]
)

nitpicky = False

# -- Options for HTML output -------------------------------------------------

html_theme = "sphinx_rtd_theme"
html_show_sphinx = False
html_static_path = ["_static"]
html_css_files = [
    str(Path("css", "rtd_width.css")),
]
# html_logo = None
# html_favicon = None

# -- Extension configuration -------------------------------------------------

todo_include_todos = True
autosectionlabel_prefix_document = True
autodoc_default_options = {
    "members": True,
    "undoc-members": True,
    "inherited-members": True,
    "autodoc_typehints": "description",
}
autoapi_type = "python"
autoapi_dirs = [str(Path(__file__).parents[1].joinpath("src"))]
autoapi_root = "api"
mermaid_version = "8.7.0"



"""
@brief Loader class for importing python based command dictionaries

@date Created July 9, 2018
@author R. Joseph Paetz

@bug No known bugs
"""

from fprime_gds.common.templates import cmd_template

# Custom Python Modules
from . import python_loader


class CmdPyLoader(python_loader.PythonLoader):
    """Class to load python based command dictionaries"""

    # Field names in the python module file dictionaries
    COMPONENT_FIELD = "COMPONENT"
    MNEMONIC_FIELD = "MNEMONIC"
    OP_CODE_FIELD = "OP_CODE"
    DESC_FIELD = "CMD_DESCRIPTION"
    ARGS_FIELD = "ARGUMENTS"

    def construct_dicts(self, path):
        """
        Constructs and returns python dictionaries keyed on id and name

        This function should not be called directly, instead use
        get_id_dict(path) and get_name_dict(path) defined in the base Loader
        class. For the command dictionary, the op_code is treated as the ID and
        the command mnemonic is treated as the name.

        Args:
            path: Path to the python module file dictionary to convert. This
                  should be a directory. If using a regular fprime deployment,
                  this should be a path to the events dictionary in your
                  generated folder:
                  ${GENERATED_FOLDER_LOCATION}/generated/${DEPLOYMENT}/events

        Returns:
            A tuple with two event dictionaries (python type dict):
            (id_dict, name_dict). They should have keys of the id and name
            fields respectively and the values for both should be event_template
            objects.
        """
        module_dicts = self.read_dict(path, use_superpkg=True)

        id_dict = {}
        name_dict = {}

        for cmd_dict in module_dicts:
            # Create a cmd template object
            cmd_temp = cmd_template.CmdTemplate(
                cmd_dict[self.OP_CODE_FIELD],
                cmd_dict[self.MNEMONIC_FIELD],
                cmd_dict[self.COMPONENT_FIELD],
                cmd_dict[self.ARGS_FIELD],
                cmd_dict[self.DESC_FIELD],
            )

            id_dict[cmd_dict[self.OP_CODE_FIELD]] = cmd_temp
            name_dict[cmd_dict[self.MNEMONIC_FIELD]] = cmd_temp

        return id_dict, name_dict



"""
@brief Channel Template class

Instances of this class describe a specific telemetry channel (but not a channel
reading)

@date Created July 11, 2018
@author R. Joseph Paetz

@bug No known bugs
"""

from fprime.common.models.serialize.type_base import BaseType
from fprime.common.models.serialize.type_exceptions import TypeMismatchException

from . import data_template


class ChTemplate(data_template.DataTemplate):
    """Class for channel templates that describe specific telemetry channels"""

    def __init__(
        self,
        ch_id,
        ch_name,
        comp_name,
        ch_type_obj,
        ch_fmt_str=None,
        ch_desc=None,
        low_red=None,
        low_orange=None,
        low_yellow=None,
        high_yellow=None,
        high_orange=None,
        high_red=None,
    ):
        """
        Constructor

        Args:
            ch_id: The ID of the channel being described
            ch_name: The name of the channel
            comp_name: The name of the f-prime component that produces this ch
            ch_type_obj: The channel's type as an instance of a class derived
                         from BaseType
            ch_fmt_str: (Optional) The format string for the channel
            ch_desc: (Optional) Description of the channel
            low_red: (Optional) Below this the value will be in red alert
            low_orange: (Optional) Below this the value will be in orange alert
            low_yellow: (Optional) Below this the value will be in yellow alert
            high_yellow: (Optional) Above this the value will be in yellow alert
            high_orange: (Optional) Above this the value will be in orange alert
            high_red: (Optional) Above this the value will be in red alert
        """
        super().__init__()
        # Make sure correct types are passed
        if not isinstance(ch_id, int):
            raise TypeMismatchException(int, type(ch_id))

        if not isinstance(ch_name, str):
            raise TypeMismatchException(str, type(ch_name))

        if not isinstance(comp_name, str):
            raise TypeMismatchException(str, type(comp_name))

        if not issubclass(ch_type_obj, BaseType):
            raise TypeMismatchException(BaseType, ch_type_obj)

        if ch_fmt_str is not None and not isinstance(ch_fmt_str, str):
            raise TypeMismatchException(str, type(ch_fmt_str))

        if ch_desc is not None and not isinstance(ch_desc, str):
            raise TypeMismatchException(str, type(ch_desc))

        # Initialize event internal variables
        self.id = ch_id
        self.name = ch_name
        self.comp_name = comp_name
        self.ch_desc = ch_desc
        self.ch_type_obj = ch_type_obj
        self.fmt_str = ch_fmt_str
        self.low_red = low_red
        self.low_orange = low_orange
        self.low_yellow = low_yellow
        self.high_yellow = high_yellow
        self.high_orange = high_orange
        self.high_red = high_red

    def get_full_name(self):
        """
        Get the full name of this channel

        Returns:
            The full name (component.channel) for this channel
        """
        return f"{self.comp_name}.{self.name}"

    def get_id(self):
        return self.id

    def get_name(self):
        return self.name

    def get_comp_name(self):
        return self.comp_name

    def get_ch_desc(self):
        return self.ch_desc

    def get_type_obj(self):
        return self.ch_type_obj

    def get_format_str(self):
        return self.fmt_str

    def get_low_red(self):
        return self.low_red

    def get_low_orange(self):
        return self.low_orange

    def get_low_yellow(self):
        return self.low_yellow

    def get_high_yellow(self):
        return self.high_yellow

    def get_high_orange(self):
        return self.high_orange

    def get_high_red(self):
        return self.high_red



"""
Tests the distributor

Created on Jul 10, 2020
@author: Josef Biberstein, Joseph Paetz, hpaulson
"""


from fprime_gds.common.distributor.distributor import Distributor
from fprime_gds.common.utils import config_manager


def test_distributor():
    """
    Tests the raw messages and leftover data for the distributor
    """
    config = config_manager.ConfigManager()
    config.set("types", "msg_len", "U16")

    dist = Distributor(config)

    header_1 = b"\x00\x0E\x00\x00\x00\x04"
    length_1 = 14
    desc_1 = 4
    data_1 = b"\x61\x62\x63\x64\x65\x66\x67\x68\x69\x6A"

    header_2 = b"\x00\x0A\x00\x00\x00\x01"
    length_2 = 10
    desc_2 = 1
    data_2 = b"\x41\x42\x43\x44\x45\x46"

    leftover_data = b"\x00\x0F\x00\x00\x00\x02\xFF"

    data = header_1 + data_1 + header_2 + data_2 + leftover_data

    (test_leftover, raw_msgs) = dist.parse_into_raw_msgs_api(bytearray(data))

    assert (
        test_leftover == leftover_data
    ), f"expected leftover data to be {list(leftover_data)}, but found {list(test_leftover)}"
    assert raw_msgs[0] == (
        header_1 + data_1
    ), f"expected first raw_msg to be {list(header_1 + data_1)}, but found {list(raw_msgs[0])}"
    assert raw_msgs[1] == (
        header_2 + data_2
    ), f"expected second raw_msg to be {list(header_2 + data_2)}, but found {list(raw_msgs[1])}"

    (test_len_1, test_desc_1, test_msg_1) = dist.parse_raw_msg_api(raw_msgs[0])
    (test_len_2, test_desc_2, test_msg_2) = dist.parse_raw_msg_api(raw_msgs[1])

    assert test_len_1 == length_1, f"expected 1st length to be {length_1} but found {test_len_1}"
    assert test_len_2 == length_2, f"expected 2nd length to be {length_2} but found {test_len_2}"
    assert test_desc_1 == desc_1, f"expected 1st desc to be {desc_1} but found {test_desc_1}"
    assert test_desc_2 == desc_2, f"expected 2nd desc to be {desc_2} but found {test_desc_2}"
    assert (test_msg_1 == data_1), f"expected 1st msg to be {list(data_1)} but found {list(test_msg_1)}"
    assert (test_msg_2 == data_2), f"expected 2nd msg to be {list(data_2)} but found {list(test_msg_2)}"


"""
A suite of unit tests for testing the predicates and filtering utilities the
GDS CLI uses
"""

import pytest

import fprime_gds.common.gds_cli.filtering_utils as filtering_utils
from fprime.common.models.serialize import time_type
from fprime_gds.common.data_types.event_data import EventData
from fprime_gds.common.templates.event_template import EventTemplate
from fprime_gds.common.testing_fw import predicates
from fprime_gds.common.utils.event_severity import EventSeverity

# Pytest fixtures work by reusing outer names, so disable this warning
# pylint: disable=redefined-outer-name
@pytest.fixture
def sample_template():
    return EventTemplate(
        event_id=12345,
        name="Luggage_Combination",
        component="helmet",
        args=[],  # Currently, no arguments given for simplicity's sake
        severity=EventSeverity.DIAGNOSTIC,
        format_str="I've got the same combination on mine!",
    )


@pytest.fixture
def sample_event(sample_template):
    event = EventData([], time_type.TimeType(), sample_template)
    return event


# A set of valid input filters that should accept the sample event
@pytest.mark.parametrize(
    "accepting_input_filter",
    [
        filtering_utils.id_predicate(12345),
        filtering_utils.get_id_predicate([]),
        filtering_utils.get_id_predicate([12345]),
        filtering_utils.get_id_predicate([451, 12345, 54321, 8]),
        filtering_utils.component_predicate("helmet"),
        filtering_utils.get_component_predicate([]),
        filtering_utils.get_component_predicate(["helmet"]),
        filtering_utils.get_component_predicate(["dark", "helmet", "luggage"]),
        filtering_utils.contains_search_string(""),
        # SysData string should contain ID, hex opcode, severity; only contains
        # the format string if args isn't None
        # Won't test this too thoroughly, since it's more based on SysData than
        # this implementation
        filtering_utils.contains_search_string("the same combination"),
        filtering_utils.get_search_predicate(""),
        filtering_utils.get_search_predicate("the same combination"),
    ],
)
@pytest.mark.gds_cli
def test_valid_accepting_filter(sample_event, accepting_input_filter):
    assert accepting_input_filter(sample_event)


@pytest.mark.parametrize(
    "rejecting_input_filter",
    [
        filtering_utils.id_predicate(54321),
        filtering_utils.get_id_predicate([123, 451, 54321, 12344]),
        filtering_utils.component_predicate("luggage"),
        filtering_utils.get_component_predicate(["dark", "luggage"]),
        filtering_utils.contains_search_string("won't be found"),
        filtering_utils.get_search_predicate("won't be found"),
    ],
)
@pytest.mark.gds_cli
def test_valid_rejecting_filter(sample_event, rejecting_input_filter):
    assert not rejecting_input_filter(sample_event)


@pytest.mark.gds_cli
def test_event_string_has_component(sample_event):
    component_filter = filtering_utils.contains_search_string("helmet")
    assert component_filter(sample_event)


@pytest.mark.gds_cli
def test_string_filter_valid_function_passing(sample_event):
    crazy_string = "tHaT'sAmAzInG..."

    def get_crazy_string(eventData) -> str:
        return crazy_string + str(eventData)

    default_filter = filtering_utils.contains_search_string(crazy_string)
    function_filter = filtering_utils.contains_search_string(
        crazy_string, get_crazy_string
    )
    assert not default_filter(sample_event)
    assert function_filter(sample_event)


@pytest.mark.gds_cli
def test_string_filter_case_sensitive(sample_event):
    upper_filter = filtering_utils.contains_search_string("Luggage_Combination")
    lower_filter = filtering_utils.contains_search_string("luggage_combination")
    assert upper_filter(sample_event)
    assert not lower_filter(sample_event)


@pytest.mark.gds_cli
def test_search_predicate_case_sensitive(sample_event):
    upper_filter = filtering_utils.get_search_predicate("Luggage_Combination")
    lower_filter = filtering_utils.get_search_predicate("luggage_combination")
    assert upper_filter(sample_event)
    assert not lower_filter(sample_event)


@pytest.mark.gds_cli
def test_time_greater_than_predicate():
    time1 = time_type.TimeType(seconds=1593610856, useconds=223451)
    time2 = time_type.TimeType(seconds=1593610856, useconds=223502)

    time_pred = predicates.greater_than(time1)
    assert not time_pred(time1)
    assert time_pred(time2)


@pytest.mark.gds_cli
def test_comparing_item_to_time_fails(sample_event):
    """
    Test that comparing a time to an event is problematic, since this assertion
    SHOULD always be true but is, in fact, false because we're comparing a data
    object to a time object
    """
    time = time_type.TimeType(seconds=12345)
    sample_event.time = time_type.TimeType(seconds=9999999)

    time_pred = predicates.greater_than(time)
    # This SHOULD return true, but it doesn't
    assert not time_pred(sample_event)


@pytest.mark.gds_cli
def test_comparing_item_to_converted_time_succeeds(sample_event):
    time = time_type.TimeType(seconds=12345)
    sample_event.time = time_type.TimeType(seconds=9999999)

    time_pred = predicates.greater_than(time)
    converted_time_pred = filtering_utils.time_to_data_predicate(time_pred)
    # Now, this returns the correct value
    assert converted_time_pred(sample_event)



"""
@brief Loader class for importing python based channel dictionaries

@date Created July 11, 2018
@author R. Joseph Paetz

@bug No known bugs
"""

from fprime_gds.common.templates.ch_template import ChTemplate

# Custom Python Modules
from .python_loader import PythonLoader


class ChPyLoader(PythonLoader):
    """Class to load python based telemetry channel dictionaries"""

    # Field names in the python module files (used to construct dictionaries)
    ID_FIELD = "ID"
    NAME_FIELD = "NAME"
    COMP_FIELD = "COMPONENT"
    DESC_FIELD = "CHANNEL_DESCRIPTION"
    TYPE_FIELD = "TYPE"
    FMT_STR_FIELD = "FORMAT_STRING"
    LOW_R_FIELD = "LOW_RED"
    LOW_O_FIELD = "LOW_ORANGE"
    LOW_Y_FIELD = "LOW_YELLOW"
    HIGH_Y_FIELD = "HIGH_YELLOW"
    HIGH_O_FIELD = "HIGH_ORANGE"
    HIGH_R_FIELD = "HIGH_RED"

    def construct_dicts(self, path):
        """
        Constructs and returns python dictionaries keyed on id and name

        This function should not be called directly, instead, use
        get_id_dict(path) and get_name_dict(path)

        Args:
            path: Path to the python module file dictionary to convert. This
                  should be a directory. If using a regular fprime deployment,
                  this should be a path to the events dictionary in your
                  generated folder:
                  ${GENERATED_FOLDER_LOCATION}/generated/${DEPLOYMENT}/channels

        Returns:
            A tuple with two channel dictionaries (python type dict):
            (id_dict, name_dict). The keys should be the channels' id and
            name fields respectively and the values should be ChTemplate
            objects.
        """
        # We do need it sometimes, so if we don't always set it to true, we will need to pass an arg
        module_dicts = self.read_dict(path, use_superpkg=True)

        id_dict = {}
        name_dict = {}

        for ch_dict in module_dicts:
            # Create a channel template object
            ch_temp = ChTemplate(
                ch_dict[self.ID_FIELD],
                ch_dict[self.NAME_FIELD],
                ch_dict[self.COMP_FIELD],
                ch_dict[self.TYPE_FIELD],
                ch_dict[self.FMT_STR_FIELD],
                ch_dict[self.DESC_FIELD],
                ch_dict[self.LOW_R_FIELD],
                ch_dict[self.LOW_O_FIELD],
                ch_dict[self.LOW_Y_FIELD],
                ch_dict[self.HIGH_Y_FIELD],
                ch_dict[self.HIGH_O_FIELD],
                ch_dict[self.HIGH_R_FIELD],
            )

            id_dict[ch_dict[self.ID_FIELD]] = ch_temp
            name_dict[ch_dict[self.NAME_FIELD]] = ch_temp

        return id_dict, name_dict



"""
@brief Class to store data from a specific event

@date Created July 2, 2018
@author R. Joseph Paetz

@bug No known bugs
"""

from fprime.common.models.serialize import time_type
from fprime_gds.common.data_types import sys_data
from fprime_gds.common.utils.string_util import format_string_template


class EventData(sys_data.SysData):
    """
    The EventData class stores a specific event message.
    """

    def __init__(self, event_args, event_time, event_temp):
        """
        Constructor.

        Args:
            event_args: The arguments of the event being stored. This should
                        be a tuple where each element is an object of a class
                        derived from the BaseType class with a filled in value.
                        Each element's class should match the class of the
                        corresponding argument type object in the event_temp
                        object. This can be None.
            event_time: The time the event occurred (TimeType)
            event_temp: Event template instance for this event

        Returns:
            An initialized EventData object
        """
        super().__init__()
        self.id = event_temp.get_id()
        self.args = event_args
        self.time = event_time
        self.template = event_temp

    def get_args(self):
        return self.args

    def get_severity(self):
        return self.template.get_severity()

    @staticmethod
    def get_empty_obj(event_temp):
        """
        Obtains an event object that is empty (arguments = None)

        Args:
            event_temp: (EventTemplate obj) Template describing event

        Returns:
            An EventData object with argument value of None
        """
        return EventData(None, time_type.TimeType(), event_temp)

    @staticmethod
    def get_csv_header(verbose=False):
        """
        Get the header for a csv file containing event data

        Args:
            verbose: (boolean, default=False) Indicates if header should be for
                                              regular or verbose output

        Returns:
            String version of the channel data
        """
        if verbose:
            return "Time,Raw Time,Name,ID,Severity,Args\n"
        else:
            return "Time,Name,Severity,Args\n"

    def get_str(self, time_zone=None, verbose=False, csv=False):
        """
        Convert the event data to a string

        Args:
            time_zone: (tzinfo, default=None) Timezone to print time in. If
                      time_zone=None, use local time.
            verbose: (boolean, default=False) Prints extra fields if True
            csv: (boolean, default=False) Prints each field with commas between
                                          if true

        Returns:
            String version of the event data
        """
        time_str = self.time.to_readable(time_zone)
        raw_time_str = str(self.time)
        name = self.template.get_full_name()
        severity = self.template.get_severity()
        format_str = self.template.get_format_str()

        if self.args is None:
            arg_str = "EMPTY EVENT OBJ"
        else:
            # The arguments are currently serializable objects which cannot be
            # used to fill in a format string. Convert them to values that can be
            arg_val_list = [arg_obj.val for arg_obj in self.args]

            arg_str = format_string_template(format_str, tuple(arg_val_list))

        if verbose and csv:
            return f"{time_str},{raw_time_str},{name},{self.id},{severity},{arg_str}"
        elif verbose and not csv:
            return f"{time_str}: {name} ({self.id}) {raw_time_str} {severity} : {arg_str}"
        elif not verbose and csv:
            return f"{time_str},{name},{severity},{arg_str}"
        else:
            return f"{time_str}: {name} {severity} : {arg_str}"

    def __str__(self):
        """
        Convert the event data to a string

        Returns:
            String version of the channel data
        """
        return self.get_str()



####
# Handles GDS logs in a lazy-loading way
####
import os
import flask_restful
import flask_restful.reqparse


class LogList(flask_restful.Resource):
    """A list of log files as produced by the GDS."""

    def __init__(self, logdir):
        """
        Constructor used to setup the log directory.

        :param logdir: log directory to search for logs
        """
        self.logdir = logdir

    def get(self):
        """Returns a list of log files that are available."""
        listing = os.listdir(self.logdir)
        return {"logs": [name for name in listing if name.endswith(".log")]}


class LogFile(flask_restful.Resource):
    """
    Command dictionary endpoint. Will return dictionary when hit with a GET.
    """

    def __init__(self, logdir):
        """
        Constructor used to setup the log directory.

        :param logdir: log directory to search for logs
        """
        self.logdir = logdir

    def get(self, name):
        """
        Returns the logdir.
        """
        logs = {}
        # Sanitization of path characters
        name = name.replace(os.path.sep, "_")
        full_path = os.path.join(self.logdir, name)
        if not os.path.exists(full_path):
            return ""
        offset = 0
        with open(full_path) as file_handle:
            file_handle.seek(offset)
            logs[name] = file_handle.read()
        return logs



####
# default_settings.py:
#
# Contains default setup for the F prime flask application. Specifically, it is used to pass configuration
# down to the GDS config layers, and is used to specify a dictionary and packet spec for specifying
# the event, channels, and commands setup.
#
# Note: flask configuration is all done via Python files
#
####
import os

import fprime_gds.common.utils.config_manager

# Select uploads directory and create it
uplink_dir = os.environ.get("UP_FILES_DIR", "/tmp/fprime-uplink/")
DOWNLINK_DIR = os.environ.get("DOWN_FILES_DIR", "/tmp/fprime-downlink/")

# Configuration is mostly driven from environment variables
DICTIONARY = os.environ.get("DICTIONARY", None)
ZMQ_TRANSPORT = os.environ.get("ZMQ_TRANSPORT", None)
ZMQ_TRANSPORT = None if ZMQ_TRANSPORT is None else ZMQ_TRANSPORT.split("|")
PORT = int(os.environ.get("TTS_PORT", "50050"), 0)
ADDRESS = os.environ.get("TTS_ADDR", "127.0.0.1")

LOG_DIR = os.environ.get("LOG_DIR", "/tmp/fprime-gds-logs")
SERVE_LOGS = os.environ.get("SERVE_LOGS", "YES") == "YES"
UPLOADED_UPLINK_DEST = uplink_dir
UPLOADS_DEFAULT_DEST = uplink_dir
REMOTE_SEQ_DIRECTORY = "/seq"
MAX_CONTENT_LENGTH = 32 * 1024 * 1024  # Max length of request is 32MiB

# Gds config setup
GDS_CONFIG = fprime_gds.common.utils.config_manager.ConfigManager()

for directory in [LOG_DIR, UPLOADED_UPLINK_DEST, UPLOADS_DEFAULT_DEST, DOWNLINK_DIR]:
    os.makedirs(directory, exist_ok=True)

# TODO: load real config



"""
chrono.py:

A chronologically-ordered history that relies on predicates to provide filtering, searching, and
retrieval operations. This history will re-order itself based on FSW time.

:author: koran
"""
from fprime.common.models.serialize.time_type import TimeType
from fprime_gds.common.history.history import History
from fprime_gds.common.testing_fw import predicates


class ChronologicalHistory(History):
    """
    A chronological history to support the GDS test api. This history adds support for specifying
    start with predicates and python's bracket notation.
    """

    ###########################################################################
    #   History Functions
    ###########################################################################
    def __init__(self, filter_pred=None):
        """
        Constructor used to set-up history. If the history is given a filter, it will ignore (drop)
        objects that don't satisfy the filter predicate.

        Args:
            filter_pred: an optional predicate to filter incoming data_objects
        """
        self.objects = []
        self.new_objects = []

        self.filter = predicates.always_true()
        if filter_pred is not None:
            if predicates.is_predicate(filter_pred):
                self.filter = filter_pred
            else:
                raise TypeError("The given filter was an instance of predicate.")

        self.retrieved_cursor = 0

    def data_callback(self, data, sender=None):
        """
        Data callback to push an object on the history. This callback will only add data_objects
        that satisfy the filter predicate.

        Args:
            data: object to store
            sender: unused API value
        """
        if self.filter(data):
            self.__insert_chrono(data, self.new_objects)
            index = self.__insert_chrono(data, self.objects)
            self.retrieved_cursor = min(index, self.retrieved_cursor)

    def retrieve(self, start=None):
        """
        Retrieve objects from this history. If a starting point is specified, will return a
        sub-list of all objects beginning at start to the latest object.
        Note: if no item satisfies the start predicate or the index is greater than the length of
        the history, an empty list will be returned.

        Args:
            start: optional first object to retrieve. can either be an index (int) or a predicate.
        Returns:
            a list of objects in chronological order
        """
        index = 0 if start is None else self.__get_index(start, self.objects)
        self.retrieved_cursor = self.size()
        self.new_objects.clear()
        return self.objects[index:]

    def retrieve_new(self, repeats=False):
        """
        Retrieves a chronological order of objects that haven't been accessed through retrieve or
        retrieve_new before.

        Returns:
            a list of objects in chronological order
        """
        index = self.retrieved_cursor
        self.retrieved_cursor = self.size()

        if repeats:
            self.new_objects.clear()
            return self.objects[index:]
        else:
            new = self.new_objects
            self.new_objects = []
            return new

    def clear(self, start=None):
        """
        Clears objects from history. A clear that specifies a starting point will clear the history
        such that start becomes the earliest element in the history after objects are removed. If
        the start is specified as a predicate, start will be the earliest object to satisfy the
        predicate.
        Note: if no item satisfies the start predicate or the index is greater than the length of
        the history, all items will be cleared.

        Args:
            start: start: an optional indicator for the first item to remove. Can be a predicate, a
                TimeType or an index in the ordering
        """
        index = self.__clear_list(start, self.objects)

        if len(self.objects) > 0:
            start = self.objects[0].get_time()
            self.__clear_list(start, self.new_objects)
        else:
            self.new_objects.clear()

        self.retrieved_cursor -= index
        self.retrieved_cursor = max(self.retrieved_cursor, 0)

    def size(self):
        """
        Accessor for the number of objects in the history
        Returns:
            the number of objects (int)
        """
        return len(self.objects)

    ###########################################################################
    #   Python Special Methods
    ###########################################################################
    def __len__(self):
        """
        Accessor for the number of objects in the history
        Returns:
            the number of objects
        """
        return self.size()

    def __getitem__(self, index):
        """
        __get_item__ is a special method in python that allows using brackets.
        Example: item = history[2] # this would return the second item in the history.

        Args:
            index: the index of the array to return.
        Returns:
            the item at the index specified.
        """
        return self.objects[index]

    ###########################################################################
    #   helper methods
    ###########################################################################
    @staticmethod
    def __insert_chrono(data_object, ordered):
        """
        traverses the existing order (back to front) and inserts the data object in the correct
        position chronologically.
        Args:
            data_object: an item to insert in the history. Must have a get_time() method.
            ordered: a list to insert the item into.
        Returns:
            the index that the item was inserted at (int)
        """
        for i, item in reversed(list(enumerate(ordered))):
            if item.get_time() < data_object.get_time():
                ordered.insert(i + 1, data_object)
                return i
        # If the data object is the earliest in the list or the list was empty
        ordered.insert(0, data_object)
        return 0

    def __clear_list(self, start, ordered):
        """
        finds the index that start specifies
        Args:
            start: an optional indicator for the first item to remove. Can be a predicate, a
                TimeType or an index in the ordering
            ordered: the list to clear
        Returns:
            the index in the given list that start refers to
        """
        index = len(ordered) if start is None else self.__get_index(start, ordered)
        del ordered[:index]
        return index

    @staticmethod
    def __get_index(start, ordered):
        """
        finds the index that start specifies
        Args:
            start: an indicator of a position in an order can be a predicate, a TimeType time
                stamp or an index in the ordering
            ordered: the list to clear
        Returns:
            the index in the given list that start refers to
        """
        if predicates.is_predicate(start):
            index = 0
            while index < len(ordered) and not start(ordered[index]):
                index += 1
            return index
        elif isinstance(start, TimeType):
            index = 0
            while index < len(ordered) and ordered[index].get_time() < start:
                index += 1
            return index
        else:
            return start



"""
@brief Base class for all loaders that load dictionaries of python fragments

The PythonLoader class inherits from the DictLoader base class and is intended
to be inherited by all loader classes that read dictionaries in the file system
made up of multiple python file fragments.

This Class only adds helper functions and thus does not overwrite any methods in
in the base DictLoader class. Because of this, all dictionaries returned from
this class will be empty.

@date Created July 3, 2018
@author R. Joseph Paetz

@bug No known bugs
"""

import glob
import importlib
import os
import sys

from fprime_gds.common.data_types import exceptions

# Custom Python Modules
from . import dict_loader


class PythonLoader(dict_loader.DictLoader):
    """Class to help load python file based dictionaries"""

    def read_dict(self, path, use_superpkg=False):
        """
        Reads all python modules at the given path and constructs a dict list

        This function assumes that path is a directory containing many python
        module files. Each module file has several fields with associated
        values. This function reads each file and constructs a dictionary using
        the fields and their values. These dictionaries are then compiled into
        a list that is returned.

        Args:
            path: File Path to a folder containing python modules to load
            use_superpkg: [Default=False] When true, modules will be imported
                           using both the package and superpackage qualifiers
                           (from A.B import C instead of from B import C). This
                           allows multiple dictionaries with the same package
                           name to be imported. This is especially important
                           when trying to import dictionaries from multiple
                           deployments.

        Returns:
            A list of dictionaries. Each dictionary represents one loaded module
            and for keys has the names of the fields in the file and for values
            has the values of those fields.
        """
        modules = self.import_modules(path, use_superpkg)

        module_dicts = []
        for module in modules:
            # Create a dictionary for this module's fields
            mod_dict = {
                field: getattr(module, field)
                for field in dir(module)
                if field.find("__") != 0 # Verify it is not a hidden field (doesn't start with "__")
            }

            module_dicts.append(mod_dict)

        return module_dicts

    @staticmethod
    def import_modules(path, use_superpkg):
        """
        Imports all modules in the given directory.

        Args:
            path: File Path to a folder containing python modules to load
            use_superpkg: When true, modules will be imported
                           using both the package and superpackage qualifiers
                           (from A.B import C instead of from B import C). This
                           allows multiple dictionaries with the same package
                           name to be imported. This is especially important
                           when trying to import dictionaries from multiple
                           deployments.

        Returns:
            A list of module objects of all the newly imported modules
        """
        # Verify path is a directory
        if not os.path.isdir(path):
            raise exceptions.GseControllerUndefinedDirectoryException(path)

        # Compute package and superpackage names
        (superpkg_path, pkg) = os.path.split(path)
        (rest_of_path, superpkg) = os.path.split(superpkg_path)

        # Make sure the directory we are importing from is in the python path
        if use_superpkg:
            sys.path.append(rest_of_path)
        else:
            sys.path.append(superpkg_path)

        # Make sure serializable directory is imported
        sys.path.append(superpkg_path + os.sep + "serializable")

        # Compute a list of all files to import
        all_files = glob.glob(path + os.sep + "*.py")

        module_files = []
        for py_file in all_files:
            (file_path, file_name) = os.path.split(py_file)
            if file_name != "__init__.py":
                module_files.append(file_name)

        # Import modules
        module_list = []
        for mf in module_files:
            # Strip off .py from name by splitting at '.' and taking the first
            # string
            mod_name = mf.split(".")[0]

            if use_superpkg:
                import_name = f"{superpkg}.{pkg}.{mod_name}"
            else:
                import_name = f"{pkg}.{mod_name}"

            m = importlib.import_module(import_name)

            module_list.append(m)

        return module_list



"""
handlers.py:

Defines a set of base classes that allow for the system to handle various functions of the system. Primarily this
defines the "DataHandler" base class for handling data.

@author mstarch
"""
import abc


class DataHandler(abc.ABC):
    """
    Defines the necessary functions required to handle data as part of the F prime project. This allows any implementer
    to be used to handle data.
    """

    @abc.abstractmethod
    def data_callback(self, data, sender=None):
        """
        Callback function used to handle data being produced elsewhere in the system and processed by the given object.
        Data supplied should be of a known type for the given object, and sender is an id of the sender. If not supplied
        sender will be None.

        :param data: data to be handled by this class
        :param sender: (optional) id of sender, otherwise None
        """


class HandlerRegistrar(abc.ABC):
    """
    Defines a class that will take in registrants and remember them for calling back later. These objects should be of
    the type "DataHandler" as this handler will send data back to these handlers when asked to do so.
    """

    def __init__(self):
        """
        Constructor defining the internal lists needed to store the registrants.
        """
        super().__init__()
        self._registrants = []

    def register(self, registrant):
        """
        Register a registrant with this registrar. Will be stored and called back when asked to send data to all the
        handlers registered.

        :param registrant: handler to register
        """
        if not isinstance(registrant, DataHandler):
            raise ValueError("Cannot register non data handler")
        self._registrants.append(registrant)

    def deregister(self, registrant):
        """
        Remove a registrant from the registrar such that it will not be called back later. Note: ignores invalid
        removals by trapping the error, as the desired effect is already satisfied.

        :param registrant: registrant to remove
        :return: True if found, False if not. May safely be ignored.
        """
        try:
            self._registrants.remove(registrant)
            return True
        except ValueError:
            return False

    def send_to_all(self, data, sender=None):
        """
        Sends the given data to all registrants.

        :param data: data to send back to registrants
        :param sender: (optional) sender to pass to data_callback
        """
        for registrant in self._registrants:
            registrant.data_callback(data, sender)



"""
@brief Decoder for file data

This decoder takes in serialized files, parses them, and packages the results
in file_data objects.

@date Created June 12, 2019
@author Blake A. Harriman

@bug No known bugs
"""

import struct

from fprime_gds.common.data_types import file_data
from fprime_gds.common.data_types.file_data import FilePacketType
from fprime_gds.common.decoders import decoder

# More info about File packets can be found at fprime-sw/Fw/FilePacket/docs/sdd.md
# Enum class for Packet Types


# Main FileDecoder class
class FileDecoder(decoder.Decoder):
    """Decoder class for file data"""

    def decode_api(self, data):
        """
        Decodes the given data and returns the result.

        This function allows for non-registered code to call the same decoding
        code as is used to parse data passed to the data_callback function.

        Args:
            data: Binary data to decode (array of bytes)

        Returns:
            Parsed version of the file data in the form of a File object
            or None if the data is not decodable
        """

        # Decode file here
        packetTypeNum, seqID = struct.unpack_from(">BI", data, 0)
        packetType = FilePacketType(packetTypeNum).name

        # Packet Type determines the variables following the seqID
        if packetType == "START":  # Packet Type is START
            fileSize, sourcePathSize = struct.unpack_from(">IB", data, 5)
            sourcePath = data[10 : sourcePathSize + 10]
            (destPathSize,) = struct.unpack_from(">B", data, sourcePathSize + 10)
            destPath = data[sourcePathSize + 11 : sourcePathSize + destPathSize + 11]
            return file_data.StartPacketData(seqID, fileSize, sourcePath, destPath)
        elif packetType == "DATA":  # Packet Type is DATA
            offset, length = struct.unpack_from(">IH", data, 5)
            dataVar = data[11 : 11 + length]
            return file_data.DataPacketData(seqID, offset, dataVar)
        elif packetType == "END":  # Packet Type is END
            hashValue = struct.unpack_from(">I", data, 5)
            return file_data.EndPacketData(seqID, hashValue)
        elif packetType == "CANCEL":  # Packet Type is CANCEL
            # CANCEL Packets have no data
            return file_data.CancelPacketData(seqID)
        # The data was not determined to be any of the packet types so return none
        return None



"""
Handles executing the "events" CLI command for the GDS
"""

from typing import Iterable

import fprime_gds.common.gds_cli.test_api_utils as test_api_utils
from fprime_gds.common.data_types.event_data import EventData
from fprime_gds.common.gds_cli.base_commands import QueryHistoryCommand
from fprime_gds.common.pipeline.dictionaries import Dictionaries
from fprime_gds.common.testing_fw import predicates
from fprime_gds.common.testing_fw.api import IntegrationTestAPI


class EventsCommand(QueryHistoryCommand):
    """
    Takes in the given arguments and prints an appropriate formatted string of
    recent event data that matches the user's criteria.

    For descriptions of these arguments, and more function details, see the
    events command definition at:
        Gds/src/fprime_gds/executables/fprime_cli.py
    """

    @classmethod
    def _get_item_list(
        cls,
        project_dictionary: Dictionaries,
        filter_predicate: predicates.predicate,
    ) -> Iterable[EventData]:
        """
        Gets a list of possible events that can occur in the system and returns
        their details in an ID-sorted list.

        :param project_dictionary: The dictionary object for the project
            containing the event type definitions
        :param filter_predicate: Test API predicate used to filter shown events
        :return: List of EventData items that passed the filter
        """
        event_list = test_api_utils.get_item_list(
            item_dictionary=project_dictionary.event_id,
            search_filter=filter_predicate,
            template_to_data=EventData.get_empty_obj,
        )
        return event_list

    @classmethod
    def _get_upcoming_item(
        cls,
        api: IntegrationTestAPI,
        filter_predicate: predicates.predicate,
        min_start_time="NOW",
        timeout: float = 5,
    ) -> EventData:
        """
        Retrieves an F' item that's occurred since the given time and returns
        its data.
        """
        return test_api_utils.get_upcoming_event(
            api, filter_predicate, min_start_time, timeout
        )



"""
flask/components.py:

This sets up the primary data components that allow Flask to connect into the system. This is where the standard
pipeline and other components are created to interact with Flask.
"""
import os
import sys

import fprime_gds.common.pipeline.standard
from fprime_gds.common.history.ram import SelfCleaningRamHistory

try:
    from fprime_gds.common.zmq_transport import ZmqClient
except ImportError:
    ZmqClient = None
from fprime_gds.common.transport import ThreadedTCPSocketClient

# Module variables, should remain hidden. These are singleton top-level objects used by Flask, and its various
# blueprints needed to run the system.
__PIPELINE = None


class FlaskEndpointRamHistory(SelfCleaningRamHistory):
    """A RAM history that is designed to work with the Flask/HTTP endpoints

    Counts object as the objects are submitted into the history tracking a total number of object tracked. During
    retrieval of the object, a counter value is saved w.r.t. the session token such that the response may be augmented
    with a validation count of objects consistent with this request. This is designed to help track GDS performance
    and track missing/lost data.

    In addition, this history creates new session tokens and uses them when a session token has not been supplied. In
    this way, HTTP clients are not free to create (potentially colliding) session tokens.
    """

    def __init__(self):
        """Constructor"""
        super().__init__()
        self.count = 0
        self.count_offsets = {}
        self.count_values = {}

    def data_callback(self, data, sender=None):
        """Counts this object before adding to history

        Callback passed in data. This variant counts the object that is supplied and passes the object back out to the
        parent history in order to store this object in the history.

        :param data: object to store
        """
        with self.lock:
            self.count += 1
            super().data_callback(data, sender)

    def retrieve(self, start=None, limit=None):
        """Retrieve objects and store current count

        Stash the current count of objects w.r.t the supplied session when supplied and then retrieve the objects in
        history through the parent. In this way the current count of objects w.r.t. the session is kept consistent with
        the last call through this retrieve function.

        Args:
            start: session key for retrieving new results
            limit: limit (count) for maximum results returned
        """
        with self.lock:

            if start is not None:
                if start not in self.retrieved_cursors:
                    self.count_offsets[start] = self.count
                self.count_values[start] = self.count - self.count_offsets[start]
            return super().retrieve(start, limit)

    def get_seen_count(self, start=None):
        """Get the count of the seen items at the time of the last retrieve call"""
        with self.lock:
            return self.count_values.get(start, self.count)


def setup_pipelined_components(
    debug,
    logger,
    config,
    dictionary,
    down_store,
    log_dir,
    tts_address,
    tts_port,
    zmq_transport,
):
    """
    Setup the standard pipeline and related components. This is done once, and then the resulting singletons are
    returned so that one object is used throughout the system.

    :param debug: used to prevent the construction of the standard pipeline
    :param logger: logger to use for output
    :param config: GDS configuration
    :param dictionary: path to F prime dictionary
    :param down_store:
    :param log_dir: log directory to write logs to, and serve logs from
    :param tts_address: address to the middleware layer
    :param tts_port: port of the middleware layer
    :param zmq_transport: set to zmq transport if using ZMQ or None to use TTS
    :return: F prime pipeline
    """
    global __PIPELINE
    if (
        __PIPELINE is None
        and not debug
        or os.environ.get("WERKZEUG_RUN_MAIN") == "true"
    ):
        if zmq_transport is not None and ZmqClient is None:
            print("[ERROR] ZMQ transportation is not available.", file=sys.stderr)
            sys.exit(-1)
        connection_uri = (
            zmq_transport
            if zmq_transport is not None
            else f"tcp://{ tts_address }:{ tts_port }"
        )
        pipeline = fprime_gds.common.pipeline.standard.StandardPipeline()
        pipeline.histories.implementation = FlaskEndpointRamHistory
        pipeline.transport_implementation = (
            ZmqClient if zmq_transport is not None else ThreadedTCPSocketClient
        )
        pipeline.setup(config, dictionary, down_store, logging_prefix=log_dir)

        logger.info(
            f"Connecting to GDS at: { connection_uri } from pid: { os.getpid() }"
        )
        pipeline.connect(connection_uri)
        __PIPELINE = pipeline
    assert __PIPELINE is not None, "Main thread did not setup pipeline appropriately"
    return __PIPELINE


def get_pipelined_components():
    """
    Returns the setup pipelined components, or raises exception if not setup yet.

    :return: F prime pipeline
    """
    assert __PIPELINE is not None, "Pipeline must be setup before use"
    return __PIPELINE



"""
base.py:

This file specifies the base-adapter for the F prime comm-layer. This class defines the basic methods needed to interact
with various wire formats as they are supported by the F prime comm-layer. This file defines a single abstract base
class representing the core features of the adapter class that must be implemented by every implementation of the
adapter for use with the comm-layer.

@author lestarch
"""
import abc


class BaseAdapter(abc.ABC):
    """
    Base adapter for adapting the communications layer. This essentially breaks down to providing the ability to read
    data from, and write to the necessary wire-format. The children of this class must at least implement the 'read' and
    'write' functions to ensure that data can be read and written. 'open' and 'close' are also provided as a helper to
    the subclass implementer to place resource initialization and release code, however; these implementations are
    defaulted not overridden.
    """

    def open(self):
        """Null default implementation"""

    def close(self):
        """Null default implementation"""

    @abc.abstractmethod
    def read(self, timeout=0.500):
        """
        Read from the interface. Must be overridden by the child adapter. Throw no fatal errors, reconnect instead. This
        call is expected to block waiting on incoming data.

        :param size: maximum size of data to read before breaking
        :param timeout: timeout for the block, default: 0.500 (500ms) as blocking w/o timeout may be uninterruptible
        :return: byte array of data, or b'' if no data was read
        """

    @abc.abstractmethod
    def write(self, frame):
        """
        Write to the interface. Must be overridden by the child adapter. Throw no fatal errors, reconnect instead.

        :param frame: framed data to uplink
        :return: True if data sent through adapter, False otherwise
        """

    @classmethod
    @abc.abstractmethod
    def get_arguments(cls):
        """
        Returns a set of arguments consumed by this adapter. This will be consumed by the CLI layer in order to provide
        command line arguments to the user. Note: these should be globally unique args, e.g. --ip-address

        :return: dictionary, keys of tuple of arg flags and value of list of other arguments to argparse's add_argument
        """

    @classmethod
    @abc.abstractmethod
    def check_arguments(cls, args):
        """
        Code that should check arguments of this adapter. If there is a problem with this code, then a "ValueError"
        should be raised describing the problem with these arguments.

        :param args: arguments as dictionary
        """

    @classmethod
    def get_adapters(cls):
        """
        Get all known adapters of this base class. These must be imported into the comm-layer to be available to the
        system, however; they only need to be imported. Once imported this function will find them and make them
        available to the comm-layer in the standard way.

        :return: list of all imported comm adapters.
        """
        adapter_map = {}
        for adapter in cls.__subclasses__():
            # Get two versions of names
            adapter_name = adapter.__module__
            adapter_short = adapter_name.split(".")[-1]
            # Check to use long or short name
            if adapter_short not in adapter_map:
                adapter_name = adapter_short
            adapter_map[adapter_name] = adapter
        return adapter_map

    @staticmethod
    def process_arguments(clazz, args):
        """
        Process arguments incoming from the command line and construct a dictionary of kwargs to supply to the chosen
        adapter at creation time. This will allow the user to choose any imported adapters at runtime.

        :param args: arguments to process
        :return: dictionary of constructor keyword arguments
        """
        return {
            value["dest"]: getattr(args, value["dest"])
            for value in clazz.get_arguments().values()
        }

    @classmethod
    def construct_adapter(cls, adapter_name, args):
        """
        Constructs a new adapter, from the given adapter name and the given namespace of argument inputs. This is a
        wrapper of "get_adapters" and "process_arguments" to help build a new, fully configured, adapter. This is a
        factory method.

        :param adapter_name: name of the adapter to build
        :param args: namespace of arg value to help build an adapter
        :return: newly constructed adapter
        """
        if adapter_name == "none":
            return None
        adapter = cls.get_adapters()[adapter_name]
        return adapter(**cls.process_arguments(adapter, args))



"""
uart.py:

This is the adapter for projects that would choose to use a serial UART interface for sending data from an F prime
deployment. This handles sending and receiving data from the things like 'LinuxSerialDriver' and other standard UART
drivers.

@author lestarch
"""

import logging

import fprime_gds.common.communication.adapters.base

import serial
from serial.tools import list_ports


LOGGER = logging.getLogger("serial_adapter")


class SerialAdapter(fprime_gds.common.communication.adapters.base.BaseAdapter):
    """
    Supplies a data source adapter that is pulling data off from a UART wire using PySerial. This is setup using a
    device handle and a baudrate for the given serial device.
    """

    BAUDS = [
        50,
        75,
        110,
        134,
        150,
        200,
        300,
        600,
        1200,
        1800,
        2400,
        4800,
        9600,
        19200,
        38400,
        57600,
        115200,
        230400,
        460800,
        500000,
        576000,
        921600,
        1000000,
        1152000,
        1500000,
        2000000,
        2500000,
        3000000,
        3500000,
        4000000,
    ]

    def __init__(self, device, baud):
        """
        Initialize the serial adapter using the default settings. This does not open the serial port, but sets up all
        the internal variables used when opening the device.
        """
        self.device = device
        self.baud = baud
        self.serial = None

    def __repr__(self):
        """ String representation for logging """
        return f"UART@{self.device}:{self.baud}bps"

    def open(self):
        """
        Opens the serial port based on previously supplied settings. If the port is already open, then close it first.
        Then open the port up again.
        """
        self.close()
        self.serial = serial.Serial(self.device, self.baud)
        return self.serial is not None

    def close(self):
        """
        Close the serial device, and ignore any errors that might arrive when attempting that closure.
        """
        try:
            if self.serial is not None:
                self.serial.close()
        finally:
            self.serial = None

    def write(self, frame):
        """
        Send a given framed bit of data by sending it out the serial interface. It will attempt to reconnect if there is
        was a problem previously. This function will return true on success, or false on error.

        :param frame: framed data packet to send out
        :return: True, when data was sent through the UART. False otherwise.
        """
        try:
            if self.serial is None:
                self.open()
            written = self.serial.write(frame)
            # Not believed to be possible to not send everything without getting a timeout exception
            assert written == len(frame)
            return True
        except serial.serialutil.SerialException as exc:
            LOGGER.warning("Serial exception caught: %s. Reconnecting.", (str(exc)))
            self.close()
        return False

    def read(self, timeout=0.500):
        """
        Read up to a given count in bytes from the UART adapter. This may return less than the full requested size but
        is expected to return some data.

        :param timeout: timeout for reading data from the serial.
        :return: data successfully read
        """
        data = b""
        try:
            if self.serial is None:
                self.open()
            # Read as much data as possible, while ensuring to block if no data is available at this time. Note: as much
            # data is read as possible to avoid a long-return time to this call. Minimum data to read is one byte in
            # order to block this function while data is incoming.
            self.serial.timeout = timeout
            data = self.serial.read(1)  # Force a block for at least 1 character
            while self.serial.in_waiting:
                data += self.serial.read(
                    self.serial.in_waiting
                )  # Drain the incoming data queue
        except serial.serialutil.SerialException as exc:
            LOGGER.warning("Serial exception caught: %s. Reconnecting.", (str(exc)))
            self.close()
        return data

    @classmethod
    def get_arguments(cls):
        """
        Returns a dictionary of flag to argparse-argument dictionaries for use with argparse to setup arguments.

        :return: dictionary of flag to argparse arguments for use with argparse
        """
        available = list(
            map(lambda info: info.device, list_ports.comports(include_links=True))
        )
        default = available[-1] if available else "/dev/ttyACM0"
        return {
            ("--uart-device",): {
                "dest": "device",
                "type": str,
                "default": default,
                "help": "UART device representing the FSW. Default: %(default)s",
            },
            ("--uart-baud",): {
                "dest": "baud",
                "type": int,
                "default": 9600,
                "help": "Baud rate of the serial device. Default: %(default)s",
            },
        }

    @classmethod
    def check_arguments(cls, args):
        """
        Code that should check arguments of this adapter. If there is a problem with this code, then a "ValueError"
        should be raised describing the problem with these arguments.

        :param args: arguments as dictionary
        """
        ports = map(lambda info: info.device, list_ports.comports(include_links=True))
        if not args["device"] in ports:
            raise ValueError(
                f"Serial port '{args['device']}' not valid. Available ports: {ports}"
            )
        # Note: baud rate may not *always* work. These are a superset
        try:
            baud = int(args["baud"])
        except ValueError:
            raise ValueError(
                f"""Serial baud rate '{args["baud"]}' not integer. Use one of: {SerialAdapter.BAUDS}"""
            )
        if not int(baud) in SerialAdapter.BAUDS:
            raise ValueError(
                f"Serial baud rate '{baud}' not supported. Use one of: {SerialAdapter.BAUDS}"
            )



"""
Tests the channel encoder

Created on Jul 10, 2020
@author: Joseph Paetz, hpaulson
"""


from fprime_gds.common.encoders.ch_encoder import ChEncoder
from fprime_gds.common.templates.ch_template import ChTemplate
from fprime_gds.common.utils.config_manager import ConfigManager
from fprime.common.models.serialize.time_type import TimeType
from fprime.common.models.serialize.numerical_types import U16Type, U32Type
from fprime_gds.common.data_types.ch_data import ChData


def test_ch_encoder():
    """
    Tests the encoding of the channel encoder
    """
    config = ConfigManager()
    config.set("types", "msg_len", "U16")

    enc = ChEncoder()
    enc_config = ChEncoder(config)

    temp = ChTemplate(101, "test_ch", "test_comp", U32Type)

    time_obj = TimeType(2, 0, 1533758629, 123456)

    ch_obj = ChData(U32Type(42), time_obj, temp)

    desc_bin = b"\x00\x00\x00\x01"
    id_bin = b"\x00\x00\x00\x65"
    time_bin = b"\x00\x02\x00\x5b\x6b\x4c\xa5\x00\x01\xe2\x40"
    val_bin = b"\x00\x00\x00\x2a"
    long_len_bin = b"\x00\x00\x00\x17"
    short_len_bin = b"\x00\x17"

    reg_expected = long_len_bin + desc_bin + id_bin + time_bin + val_bin
    config_expected = short_len_bin + desc_bin + id_bin + time_bin + val_bin

    reg_output = enc.encode_api(ch_obj)

    assert (
        reg_output == reg_expected
    ), f"FAIL: expected regular output to be {list(reg_expected)}, but found {list(reg_output)}"

    config_output = enc_config.encode_api(ch_obj)

    assert (
        config_output == config_expected
    ), f"FAIL: expected configured output to be {list(config_expected)}, but found {list(config_output)}"

    temp = ChTemplate(102, "test_ch2", "test_comp2", U16Type)

    time_obj = TimeType(2, 0, 1533758628, 123457)

    ch_obj = ChData(U16Type(40), time_obj, temp)

    desc_bin = b"\x00\x00\x00\x01"
    id_bin = b"\x00\x00\x00\x66"
    time_bin = b"\x00\x02\x00\x5b\x6b\x4c\xa4\x00\x01\xe2\x41"
    val_bin = b"\x00\x28"
    long_len_bin = b"\x00\x00\x00\x15"
    short_len_bin = b"\x00\x15"

    reg_expected = long_len_bin + desc_bin + id_bin + time_bin + val_bin
    config_expected = short_len_bin + desc_bin + id_bin + time_bin + val_bin

    reg_output = enc.encode_api(ch_obj)

    assert (
        reg_output == reg_expected
    ), f"FAIL: expected regular output to be {list(reg_expected)}, but found {list(reg_output)}"

    config_output = enc_config.encode_api(ch_obj)

    assert (
        config_output == config_expected
    ), f"FAIL: expected configured output to be {list(config_expected)}, but found {list(config_output)}"



"""
Tests the packet encoder

Created on Jul 10, 2020
@author: Joseph Paetz, hpaulson
"""


from fprime_gds.common.encoders.pkt_encoder import PktEncoder
from fprime_gds.common.data_types.pkt_data import PktData
from fprime_gds.common.utils.config_manager import ConfigManager
from fprime_gds.common.templates.ch_template import ChTemplate
from fprime_gds.common.templates.pkt_template import PktTemplate
from fprime_gds.common.data_types.ch_data import ChData
from fprime.common.models.serialize.time_type import TimeType
from fprime.common.models.serialize.numerical_types import U8Type, U16Type, U32Type


def test_pkt_encoder():
    """
    Tests the encoding of the packet encoder
    """
    config = ConfigManager()
    config.set("types", "msg_len", "U16")

    enc = PktEncoder()
    enc_config = PktEncoder(config)

    ch_temp_1 = ChTemplate(101, "test_ch", "test_comp", U32Type)
    ch_temp_2 = ChTemplate(102, "test_ch2", "test_comp2", U8Type)
    ch_temp_3 = ChTemplate(103, "test_ch3", "test_comp3", U16Type)

    pkt_temp = PktTemplate(64, "test_pkt", [ch_temp_1, ch_temp_2, ch_temp_3])

    time_obj = TimeType(2, 0, 1533758629, 123456)

    ch_obj_1 = ChData(U32Type(1356), time_obj, ch_temp_1)
    ch_obj_2 = ChData(U8Type(143), time_obj, ch_temp_2)
    ch_obj_3 = ChData(U16Type(1509), time_obj, ch_temp_3)

    pkt_obj = PktData([ch_obj_1, ch_obj_2, ch_obj_3], time_obj, pkt_temp)

    desc_bin = b"\x00\x00\x00\x04"
    id_bin = b"\x00\x40"
    time_bin = b"\x00\x02\x00\x5b\x6b\x4c\xa5\x00\x01\xe2\x40"
    ch_bin = b"\x00\x00\x05\x4c\x8F\x05\xe5"
    long_len_bin = b"\x00\x00\x00\x18"
    short_len_bin = b"\x00\x18"

    reg_expected = long_len_bin + desc_bin + id_bin + time_bin + ch_bin
    config_expected = short_len_bin + desc_bin + id_bin + time_bin + ch_bin

    reg_output = enc.encode_api(pkt_obj)

    assert (
        reg_output == reg_expected
    ), f"FAIL: expected regular output to be {list(reg_expected)}, but found {list(reg_output)}"

    config_output = enc_config.encode_api(pkt_obj)

    assert (
        config_output == config_expected
    ), f"FAIL: expected configured output to be {list(config_expected)}, but found {list(config_output)}"



""" stats.py:

Statistics package for the GDS to help diagnose performance issues and give greater visibility into the running GDS.
"""
import flask_restful

from typing import Dict
from fprime_gds.common.history.history import History
from fprime_gds.common.history.ram import RamHistory


class StatsBlob(flask_restful.Resource):
    """Produces a statistics blob for the upstream webpage"""

    def __init__(self, histories: Dict[str, History]):
        """Construct the stats object with knowledge of the histories"""
        self.histories = histories

    def get(self):
        """Get method for the statistics package"""
        try:
            counts = {
                key: history.sessions()
                for key, history in self.histories.items()
                if isinstance(history, RamHistory)
            }
            sizes = {key: history.size() for key, history in self.histories.items()}
            return {
                "Active Clients": {"total": max(counts.values()), **counts},
                "History Sizes": {"total": sum(sizes.values()), **sizes},
            }
        except Exception as exc:
            return {"errors": [str(exc)]}



####
# run_deployment.py:
#
# Runs a deployment. Starts a GUI, a TCPServer, and the deployment application.
####
import os
import sys
import platform
import webbrowser
from pathlib import Path
from fprime.fbuild.settings import (
    IniSettings,
    FprimeLocationUnknownException,
    FprimeSettingsException,
)

import fprime_gds.executables.cli
import fprime_gds.executables.utils


def get_artifacts_root() -> Path:
    try:
        ini_settings = IniSettings.load(None)
    except FprimeLocationUnknownException:
        print(
            "[ERROR] Not in fprime deployment and no artifacts root provided, unable to find dictionary and/or app"
        )
        sys.exit(-1)
    except FprimeSettingsException as e:
        print("[ERROR]", e)
        sys.exit(-1)
    assert "install_dest" in ini_settings, "install_dest not in settings.ini"
    print(
        f"""[INFO] Autodetected artifacts root '{ini_settings["install_dest"]}' from deployment settings.ini file."""
    )
    return ini_settings["install_dest"]


def find_app(root: Path) -> Path:
    bin_dir = root / "bin"

    if not bin_dir.exists():
        print(f"[ERROR] binary location {bin_dir} does not exist")
        sys.exit(-1)

    files = []
    for child in bin_dir.iterdir():
        if child.is_file():
            files.append(child)

    if len(files) == 0:
        print(f"[ERROR] App not found in {bin_dir}")
        sys.exit(-1)

    if len(files) > 1:
        print(
            f"[ERROR] Multiple app candidates in binary location {bin_dir}. Specify app manually with --app."
        )
        sys.exit(-1)

    return files[0]


def find_dict(root: Path) -> Path:
    dict_dir = root / "dict"

    if not dict_dir.exists():
        print(f"[ERROR] dictionary location {dict_dir} does not exist")
        sys.exit(-1)

    files = []
    for child in dict_dir.iterdir():
        if child.is_file() and child.suffix == ".xml":
            files.append(child)

    if len(files) == 0:
        print(f"[ERROR] No xml dictionary found in dictionary location {dict_dir}")
        sys.exit(-1)

    if len(files) > 1:
        print(
            f"[ERROR] Multiple xml dictionaries found in dictionary location {dict_dir}. Specify dictionary manually with --dictionary."
        )
        sys.exit(-1)

    return files[0]


def get_settings():
    args = parse_args()

    if args.dictionary is not None and (args.app is not None or args.noapp):
        return args

    root = None
    if args.root_dir is not None:
        root = Path(args.root_dir)
    else:
        root = get_artifacts_root()
    root = root / platform.system()

    if not args.noapp and args.app is None:
        args.app = find_app(root)

    if args.dictionary is None:
        args.dictionary = find_dict(root)

    return args


def parse_args():
    """
    Gets an argument parsers to read the command line and process the arguments. Return
    the arguments in their namespace.

    :return: parsed argument namespace
    """
    # Get custom handlers for all executables we are running
    arg_handlers = [
        fprime_gds.executables.cli.LogDeployParser,
        fprime_gds.executables.cli.GdsParser,
        fprime_gds.executables.cli.MiddleWareParser,
        fprime_gds.executables.cli.BinaryDeployment,
        fprime_gds.executables.cli.CommParser,
    ]
    # Parse the arguments, and refine through all handlers
    try:
        args, parser = fprime_gds.executables.cli.ParserBase.parse_args(
            arg_handlers, "Run F prime deployment and GDS"
        )
        # Special checks
        if args.config.get_file_path() is None and args.gui == "wx":
            raise ValueError("Must supply --config when using 'wx' GUI.")
    # On ValueError print error, help and exit
    except ValueError as vexc:
        print(f"[ERROR] {str(vexc)}", file=sys.stderr, end="\n\n")
        parser.print_help(sys.stderr)
        sys.exit(-1)
    return args


def launch_process(cmd, logfile=None, name=None, env=None, launch_time=5):
    """
    Launch a child subprocess. This subprocess will allow the child to run outside of the memory context of Python.

    :param cmd: list of command arguments to run by handing to subprocess.
    :param logfile: (optional) place to redirect output to for purposes of logging. Default: None, screen.
    :param name: (optional) short name for printing messages.
    :param env: (optional) environment to run in. Allows for special environment contexts.
    :param launch_time: (optional) time to launch the process, before rendering an error.
    :return: running process
    """
    if name is None:
        name = str(cmd)
    print(f"[INFO] Ensuring {name} is stable for at least {launch_time} seconds")
    try:
        return fprime_gds.executables.utils.run_wrapped_application(
            cmd, logfile, env, launch_time
        )
    except fprime_gds.executables.utils.AppWrapperException as awe:
        print(f"[ERROR] {str(awe)}.", file=sys.stderr)
        try:
            if logfile is not None:
                with open(logfile) as file_handle:
                    for line in file_handle.readlines():
                        print(f"    [LOG] {line.strip()}", file=sys.stderr)
        except Exception:
            pass
        raise fprime_gds.executables.utils.AppWrapperException(f"Failed to run {name}")


def launch_tts(tts_port, tts_addr, logs, **_):
    """
    Launch the Threaded TCP Server

    :param tts_port: port to attach to
    :param tts_addr: address to bind to
    :param logs: logs output directory
    :return: process
    """
    # Open log, and prepare to close it cleanly on exit
    tts_log = os.path.join(logs, "ThreadedTCP.log")
    # Launch the tcp server
    tts_cmd = [
        sys.executable,
        "-u",
        "-m",
        "fprime_gds.executables.tcpserver",
        "--port",
        str(tts_port),
        "--host",
        str(tts_addr),
    ]
    return launch_process(tts_cmd, logfile=tts_log, name="TCP Server")


def launch_wx(port, dictionary, connect_address, log_dir, config, **_):
    """
    Launch the GDS gui

    :param port: port to connect to
    :param dictionary: dictionary to look at
    :param connect_address: address to connect to
    :param log_dir: directory to place logs
    :param config: configuration to use
    :return: process
    """
    gse_args = [
        sys.executable,
        "-u",
        "-m",
        "fprime_gds.wxgui.tools.gds",
        "--port",
        str(port),
    ]
    if os.path.isfile(dictionary):
        gse_args.extend(["-x", dictionary])
    elif os.path.isdir(dictionary):
        gse_args.extend(["--dictionary", dictionary])
    else:
        print(
            f"[ERROR] Dictionary invalid, must be XML or PY dicts: {dictionary}",
            file=sys.stderr,
        )
    # For macOS, add in the wx wrapper
    if platform.system() == "Darwin":
        gse_args.insert(
            0,
            os.path.join(
                os.path.dirname(__file__),
                "..",
                "..",
                "..",
                "bin",
                "osx",
                "wx-wrapper.bash",
            ),
        )
    gse_args.extend(
        ["--addr", connect_address, "-L", log_dir, "--config", config.get_file_path()]
    )
    return launch_process(gse_args, name="WX GUI")


def launch_html(
    tts_port, dictionary, connect_address, logs, gui_addr, gui_port, **extras
):
    """
    Launch the flask server and a browser pointed at the HTML page.

    :param tts_port: port to connect to
    :param dictionary: dictionary to look at
    :param connect_address: address to connect to
    :param logs: directory to place logs
    :param gui_addr: Flask server host IP address
    :param gui_port: Flask server port number
    :return: process
    """
    gse_env = os.environ.copy()
    gse_env.update(
        {
            "DICTIONARY": str(dictionary),
            "FLASK_APP": "fprime_gds.flask.app",
            "LOG_DIR": logs,
            "SERVE_LOGS": "YES",
        }
    )
    if tts_port is not None:
        gse_env.update(
            {
                "TTS_PORT": str(tts_port),
                "TTS_ADDR": connect_address,
            }
        )
    else:
        gse_env.update({"ZMQ_TRANSPORT": "|".join(connect_address)})

    gse_args = [
        sys.executable,
        "-u",
        "-m",
        "flask",
        "run",
        "--host",
        str(gui_addr),
        "--port",
        str(gui_port),
    ]
    ret = launch_process(gse_args, name="HTML GUI", env=gse_env, launch_time=2)
    if extras["gui"] == "html":
        webbrowser.open(
            f"http://{str(gui_addr)}:{str(gui_port)}/", new=0, autoraise=True
        )
    return ret


def launch_app(app, port, address, logs, **_):
    """
    Launch the app

    :param app: application to launch
    :param port: port to connect to
    :param address: address to connect to
    :param logs: log directory to place files into
    :return: process
    """
    app_name = os.path.basename(app)
    logfile = os.path.join(logs, f"{app_name}.log")
    app_cmd = [os.path.abspath(app), "-p", str(port), "-a", address]
    return launch_process(
        app_cmd, name=f"{app_name} Application", logfile=logfile, launch_time=1
    )


def launch_comm(comm_adapter, tts_port, connect_address, logs, **all_args):
    """

    :return:
    """
    transport_args = ["--tts-addr", connect_address, "--tts-port", str(tts_port)]
    if tts_port is None:
        transport_args = ["--zmq", "--zmq-transport", connect_address[0], connect_address[1], "--zmq-server"]
    app_cmd = [
        sys.executable,
        "-u",
        "-m",
        "fprime_gds.executables.comm",
        "-l",
        logs,
        "--log-directly",
        "--comm-adapter",
        all_args["adapter"],
        "--comm-checksum-type",
        all_args["checksum_type"],
    ] + transport_args
    # Manufacture arguments for the selected adapter
    for arg in comm_adapter.get_arguments().keys():
        definition = comm_adapter.get_arguments()[arg]
        destination = definition["dest"]
        app_cmd.append(arg[0])
        app_cmd.append(str(all_args[destination]))
    return launch_process(
        app_cmd, name=f'comm[{all_args["adapter"]}] Application', launch_time=1
    )


def main():
    """
    Main function used to launch processes.
    """
    settings = vars(get_settings())
    launchers = []
    # Launch a gui, if specified
    if settings["zmq"]:
        settings["connect_address"] = settings["zmq_transport"]
        settings["tts_port"] = None
    else:
        launchers.append(launch_tts)
        settings["connect_address"] = (
            settings["tts_addr"] if settings["tts_addr"] != "0.0.0.0" else "127.0.0.1"
        )
    # Check if we are running with communications
    if settings.get("adapter", "") != "none":
        launchers.append(launch_comm)

    # Add app, if possible
    if settings.get("app", None) is not None and settings.get("adapter", "") == "ip":
        launchers.append(launch_app)
    elif settings.get("app", None) is not None:
        print("[WARNING] App cannot be auto-launched without IP adapter")

    # Launch the desired GUI package
    gui = settings.get("gui", "none")
    if gui == "wx":
        launchers.append(launch_wx)
    elif gui in ["html", "none"]:
        launchers.append(launch_html)
    # elif gui == "none":
    #    print("[WARNING] No GUI specified, running headless", file=sys.stderr)
    else:
        raise Exception(f'Invalid GUI specified: {settings["gui"]}')
    # Launch launchers and wait for the last app to finish
    try:
        procs = []
        for launcher in launchers:
            procs.append(launcher(**settings))
        print("[INFO] F prime is now running. CTRL-C to shutdown all components.")
        procs[-1].wait()
    except KeyboardInterrupt:
        print("[INFO] CTRL-C received. Exiting.")
    except Exception as exc:
        print(
            f"[INFO] Shutting down F prime due to error. {str(exc)}",
            file=sys.stderr,
        )
        return 1
    # Processes are killed atexit
    return 0


if __name__ == "__main__":
    sys.exit(main())



""" resource.py: flask_restful resource helpers

Helper functions used to manage resource endpoints for the fprime_gds. This functions to reduce code duplication,
establish standard patterns, and prevent replicated errors.

@author lestarch
"""
from flask_restful import Resource
from flask_restful.reqparse import RequestParser
from fprime_gds.flask.errors import build_error_object


class DictionaryResource(Resource):
    """
    Resource tasked with serving the supplied dictionary through flask. The dictionary will be returned as-is and thus
    should be flask compatible. Errors with the dictionary are a 500 server error and may not be recovered.
    """

    def __init__(self, dictionary):
        """Constructor used to setup for dictionary

        Args:
            dictionary: dictionary to serve when GET is called
        """
        self.dictionary = dictionary

    def get(self):
        """HTTP GET method handler for dictionary resource

        Returns:
            dictionary ready for conversion into JSON
        """
        return self.dictionary


class HistoryResourceBase(Resource):
    """
    Base class for resources serving histories. Internalizes a history, serves the latest data from that history, and
    clears the history data such that replicated data is not sent. The GET method will loop through history object
    calling `process` to allow subclasses to post-process the object for sending. Errors in process will be aggregated
    but will not fail the GET transaction, however; errors outside of process will result in a 500 error.

    The history base object also sets up the request parser to handle the session argument needed to track the pointer
    into the history. This session is automatically deleted when the DELETE handler is invoked such that the GDS is
    kept cleaned-up.
    """

    def __init__(self, history):
        """Construct this history resource around a supplied history

        Args:
            history: history used as a base data store for this resource
        """
        self.parser = RequestParser()
        self.parser.add_argument(
            "session", required=True, help="Session key for fetching data."
        )
        self.parser.add_argument(
            "limit", required=False, help="Limit to results returned (default 2000)"
        )

        self.history = history

    def process(self, item):
        """Base history object processing function (does nothing)"""
        return item

    def get(self):
        """HTTP GET handler returning new history objects"""
        errors = []
        returned_items = []
        args = self.parser.parse_args()

        # Set the clear time if possible
        if hasattr(self.history, "set_clear_time"):
            self.history.set_clear_time(60)

        # Get the new items from history ensuring it is clear in a fail-safe attempt to repeat recuring errors
        try:
            session = args.get("session")
            limit = args.get("limit") if args.get("limit") is not None else 2000
            new_items = self.history.retrieve(session, int(limit))
            validation = -1
            if hasattr(self.history, "get_seen_count"):
                validation = self.history.get_seen_count(session)
        finally:
            self.history.clear()

        # Process each item from history aggregating but not failing on processing errors
        for item in new_items:
            try:
                returned_items.append(self.process(item))
            except Exception as exc:
                errors.append(build_error_object(exc))
        return {"history": returned_items, "validation": validation, "errors": errors}



####
# app.py:
#
# This file sets up the flask app, and registers the endpoints for the individual APIs supplied by
# this framework.
#
####
import logging
import os
import sys
import uuid
import flask

# Try to import Compress, but disable compression if not installed
try:
    from flask_compress import Compress
except ImportError:
    Compress = None

from fprime_gds.flask import flask_uploads

import fprime_gds.flask.channels

# Import the Flask API implementations
import fprime_gds.flask.commands
import fprime_gds.flask.events
import fprime_gds.flask.json
import fprime_gds.flask.logs
import fprime_gds.flask.updown
import fprime_gds.flask.sequence
import fprime_gds.flask.stats
import fprime_gds.flask.errors

from . import components


# Update logging to avoid redundant messages
logger = logging.getLogger("werkzeug")
logger.setLevel(logging.WARN)
logger = logging.getLogger("downlink")
logger.setLevel(logging.INFO)


def construct_app():
    """
    Constructs the Flask app by taking the following steps:

    1. Setup and configure the app
    2. Setup JSON encoding for Flask and flask_restful to handle F prime types natively
    3. Setup standard pipeline used throughout the system
    4. Create Restful API for registering flask items
    5. Setup flask_uploads settings
    6. Register all restful endpoints

    :return: setup app
    """
    app = flask.Flask(__name__, static_url_path="")
    # Enable compression if it is installed
    if Compress is not None:
        compress = Compress()
        compress.init_app(app)

    app.config.from_object("fprime_gds.flask.default_settings")
    # Override defaults from python files specified in 'FP_FLASK_SETTINGS'
    if "FP_FLASK_SETTINGS" in os.environ:
        app.config.from_envvar("FP_FLASK_SETTINGS")

    # JSON encoding setting must come before restful
    app.json_encoder = fprime_gds.flask.json.GDSJsonEncoder
    app.config["RESTFUL_JSON"] = {"cls": app.json_encoder}
    # Standard pipeline creation
    pipeline = components.setup_pipelined_components(
        app.debug,
        app.logger,
        app.config["GDS_CONFIG"],
        app.config["DICTIONARY"],
        app.config["DOWNLINK_DIR"],
        app.config["LOG_DIR"],
        app.config["ADDRESS"],
        app.config["PORT"],
        app.config["ZMQ_TRANSPORT"],
    )

    # Restful API registration
    api = fprime_gds.flask.errors.setup_error_handling(app)
    # File upload configuration, 1 set for everything
    uplink_set = flask_uploads.UploadSet("uplink", flask_uploads.ALL)
    flask_uploads.configure_uploads(app, [uplink_set])

    # Application routes
    api.add_resource(
        fprime_gds.flask.commands.CommandDictionary,
        "/dictionary/commands",
        resource_class_args=[pipeline.dictionaries.command_name],
    )
    api.add_resource(
        fprime_gds.flask.commands.CommandHistory,
        "/commands",
        resource_class_args=[pipeline.histories.commands],
    )
    api.add_resource(
        fprime_gds.flask.commands.Command,
        "/commands/<command>",
        resource_class_args=[pipeline],
    )
    api.add_resource(
        fprime_gds.flask.events.EventDictionary,
        "/dictionary/events",
        resource_class_args=[pipeline.dictionaries.event_id],
    )
    api.add_resource(
        fprime_gds.flask.events.EventHistory,
        "/events",
        resource_class_args=[pipeline.histories.events],
    )
    api.add_resource(
        fprime_gds.flask.channels.ChannelDictionary,
        "/dictionary/channels",
        resource_class_args=[pipeline.dictionaries.channel_id],
    )
    api.add_resource(
        fprime_gds.flask.channels.ChannelHistory,
        "/channels",
        resource_class_args=[pipeline.histories.channels],
    )
    api.add_resource(
        fprime_gds.flask.updown.Destination,
        "/upload/destination",
        resource_class_args=[pipeline.files.uplinker],
    )
    api.add_resource(
        fprime_gds.flask.updown.FileUploads,
        "/upload/files",
        resource_class_args=[pipeline.files.uplinker, uplink_set],
    )
    api.add_resource(
        fprime_gds.flask.updown.FileDownload,
        "/download/files",
        "/download/files/<string:source>",
        resource_class_args=[pipeline.files.downlinker],
    )
    api.add_resource(
        fprime_gds.flask.sequence.SequenceCompiler,
        "/sequence",
        resource_class_args=[
            app.config["DICTIONARY"],
            app.config["UPLOADED_UPLINK_DEST"],
            pipeline.files.uplinker,
            app.config["REMOTE_SEQ_DIRECTORY"],
        ],
    )
    api.add_resource(
        fprime_gds.flask.stats.StatsBlob,
        "/stats",
        resource_class_args=[
            {
                "events": pipeline.histories.events,
                "channels": pipeline.histories.channels,
                "commands": pipeline.histories.commands,
            }
        ],
    )

    # Optionally serve log files
    if app.config["SERVE_LOGS"]:
        api.add_resource(
            fprime_gds.flask.logs.LogList,
            "/logdata",
            resource_class_args=[app.config["LOG_DIR"]],
        )
        api.add_resource(
            fprime_gds.flask.logs.LogFile,
            "/logdata/<name>",
            resource_class_args=[app.config["LOG_DIR"]],
        )
    return app, api


try:
    app, _ = construct_app()
except Exception as exc:
    print(f"[ERROR] {exc}", file=sys.stderr)
    sys.exit(1)


@app.errorhandler(Exception)
def handle_unexpected_error(error):
    status_code = 500
    response = {"errors": [fprime_gds.flask.errors.build_error_object(error)]}
    return flask.jsonify(response), status_code


@app.route("/js/<path:path>")
def files_serve(path):
    """
    A function used to serve the JS files needed for the GUI layers.

    :param path: path to the file (in terms of web browser)
    """
    return flask.send_from_directory("static/js", path)


@app.route("/")
def index():
    """
    A function used to serve the JS files needed for the GUI layers.
    """
    return flask.send_from_directory("static", "index.html")


@app.route("/logs")
def log():
    """
    A function used to serve the JS files needed for the GUI layers.
    """
    return flask.send_from_directory("static", "logs.html")


@app.route("/session")
def session():
    return flask.jsonify({"session": uuid.uuid4()}), 200


@app.after_request
def set_no_cache(response):
    """Set the no-cache header"""
    response.headers["Cache-Control"] = "no-cache"
    return response


# When running from the command line, this will allow the flask development server to launch
if __name__ == "__main__":
    app.run()



####
# channels.py:
#
# This file captures the HTML endpoint for the events list. This file contains one endpoint which allows for listing
# events after a certain time-stamp.
#
#  GET /events: list events
#      Input Data: {
#                      "start-time": "YYYY-MM-DDTHH:MM:SS.sss" #Start time for event listing
#                  }
####
import copy
from fprime.common.models.serialize.serializable_type import SerializableType
from fprime.common.models.serialize.array_type import ArrayType
from fprime_gds.flask.resource import DictionaryResource, HistoryResourceBase


class ChannelDictionary(DictionaryResource):
    """Channel dictionary shares implementation"""


class ChannelHistory(HistoryResourceBase):
    """
    Resource supplying the history of channels in the system. Includes `get_display_text` postprocessing to add in the
    getter for the display text.
    """

    def process(self, chan):
        """Process the channel to add get_display_text"""
        chan = copy.copy(chan)
        # Setup display_text and when needed
        if isinstance(chan.val_obj, (SerializableType, ArrayType)):
            setattr(chan, "display_text", chan.val_obj.formatted_val)
        elif chan.template.get_format_str() is not None:
            setattr(
                chan,
                "display_text",
                chan.template.get_format_str() % (chan.val_obj.val),
            )
        return chan



"""
TODO: Find more appropriate places for these.

This is the Isle of Misfit Functions used by the GDS CLI, ones that I wasn't
sure where else to put. Please find a better home for them when you have time
(i.e. a more organized file).
"""

import json
from typing import Callable, Iterable, NamedTuple

from fprime_gds.common.data_types.sys_data import SysData
from fprime_gds.common.templates.cmd_template import CmdTemplate

# NOTE: 'GDSJsonEncoder' module is now lazy-loaded below for performance
# reasons


# Define some tuples containing sets of arguments used by GDS CLI commands;
# For more details on these arguments, see argument definitions at:
#           Gds/src/fprime_gds/executables/fprime_cli.py
ConnectionInfo = NamedTuple(
    "ConnectionInfo", [("dictionary", str), ("ip_address", str), ("port", int)]
)
SearchInfo = NamedTuple(
    "SearchInfo",
    [
        ("is_printing_list", bool),
        ("ids", Iterable[int]),
        ("components", Iterable[str]),
        ("search", str),
        ("json", bool),
    ],
)


def repeat_until_interrupt(func: Callable, *args):
    """
    Continues to call the input function with the given arguments until the
    user interrupts it.

    :param func: The function you want to call repeatedly. If the function
        returns anything, it MUST return a new, updated tuple of the arguments
        passed into it in the same order, which will be used as the new
        arguments in the next iteration. This is done to allow for persistent
        state between iterations; if needed, create a wrapper for your original
        function to do this. If the function does NOT return anything, the
        original input arguments will continue to be used
    :param args: All keyword arguments you want to pass into "func"
    """
    try:
        while True:
            new_args = func(*args)  # lgtm [py/call/wrong-arguments]
            if new_args:
                args = new_args
    except KeyboardInterrupt:
        pass


def get_item_json_string(gds_item, tab_spaces: int = 2) -> str:
    """
    Converts the given F Prime GDS data item into a JSON-formatted string.

    :param gds_item: The item to convert to JSON
    :param tab_spaces: The (number of spaces to use for indented lines
    :return: A string of the gds_item in JSON format
    """
    from fprime_gds.flask.json import GDSJsonEncoder

    return json.dumps(gds_item, indent=tab_spaces, cls=GDSJsonEncoder)


def get_item_string(item: SysData, as_json: bool = False) -> str:
    """
    Takes in the given SysData and prints out a human-readable string
    representation of its information.

    :param item: The item to create a string for
    :param as_json: Return a JSON-string representation of the given item
        instead
    :return: A readable string of the item information
    """
    if not item:
        return ""

    if as_json:
        return get_item_json_string(item)
    # NOTE: "get_str" isn't on the base sys_data class, but is on all the query
    # items we care about so far (i.e. EventData, ChannelData, CommandData)
    return item.get_str(verbose=True)


def get_cmd_template_string(
    item: CmdTemplate,
    as_json: bool = False,
) -> str:
    """
    Converts the given command template into a human-readable string.

    :param item: The CmdTemplate to convert to a string
    :param as_json: Whether or not to return a JSON representation of "temp"
    :return: A readable string version of "item"
    """
    if not item:
        return get_item_string(item)
    if as_json:
        return get_item_json_string(item)

    cmd_string = "%s (%d) | Takes %d arguments.\n" % (
        item.get_full_name(),
        item.get_id(),
        len(item.get_args()),
    )

    cmd_description = item.get_description()
    if cmd_description:
        cmd_string += f"Description: {(cmd_description)}\n" 

    for arg in item.get_args():
        arg_name, arg_description, arg_type = arg
        if not arg_description:
            arg_description = "--no description--"
        # Can't compare against actual module, since EnumType is a serializable
        # type from the dictionary
        if type(arg_type).__name__ == "EnumType":
            arg_description = f"{str(arg_type.keys())} | {arg_description} "

        cmd_string += f"\t{arg_name} ({type(arg_type).__name__}): {arg_description}\n"

    return cmd_string



""" File containing checksum implementations and function to calculate checksum implementation

fprime has historically supported several types of checksums. The primary is CRC32 and the constant testing-only
checksum, which was a constant.
"""
import zlib


def crc_calculation(data: bytes):
    """Initial checksum implementation for FpFramerDeframer."""
    return zlib.crc32(data) & 0xFFFFFFFF


CHECKSUM_SELECTION = "crc32"
CHECKSUM_MAPPING = {
    "fixed": lambda data: 0xCAFECAFE,
    "crc32": crc_calculation,
    "default": crc_calculation,
}


def calculate_checksum(data: bytes):
    """Calculates the checksum of bytes"""
    selected_checksum = CHECKSUM_SELECTION
    hash_fn = CHECKSUM_MAPPING.get(selected_checksum, CHECKSUM_MAPPING.get("default"))
    return hash_fn(data)



"""
@brief Base class for all encoders. Defines the Encoder interface.

Encoders are responsible for taking data objects and serializing them into
binary data that can be sent to an fprime deployment.

Producers of encodable data (such as gui panels) will be passed an encoder
object for a specific descriptor type (cmd, file, etc). This will register the
encoder with the panel and the panel is responsible for calling the encoder's
data_callback function with data of the given descriptor type to send. In turn,
a sender (such as a tcp client) will be registered to the encoder using the
encoder's register_send function. The encoder will then call the sender's send
function with the serialized binary data.

This base class does not do any serialization. Giving data to it is a no-op. Its
purpose is to define the interface for an encoder.

@date Created July 9, 2018
@author R. Joseph Paetz

@bug No known bugs
"""
import abc
import logging

import fprime_gds.common.handlers
from fprime_gds.common.utils.config_manager import ConfigManager

LOGGER = logging.getLogger("encoder")


class Encoder(
    fprime_gds.common.handlers.DataHandler,
    fprime_gds.common.handlers.HandlerRegistrar,
    abc.ABC,
):
    """
    Base class for all encoder classes. This defines the "encode_api" function to allow for decoding of raw bytes. In
    addition it has a "data_callback" function implementation that decodes and sends out all results.
    """

    def __init__(self, config=None):
        """
        Encoder class constructor

        :param config: (ConfigManager, default=None): Object with configuration data for the sizes of fields in the
                       binary data. If None passed, defaults are used.
        """
        super().__init__()
        if config is None:
            # Retrieve defaults for the configs
            config = ConfigManager()
        self.config = config

    def data_callback(self, data, sender=None):
        """
        Data callback which calls the encode_api function exactly once. Then it passes the results to all registered
        consumer. This should only need to be overridden in extraordinary circumstances.

        :param data: data bytes to be decoded
        :param sender: (optional) sender id, otherwise None
        :return: returns the encoded data for reference
        """
        encoded = self.encode_api(data)
        if encoded is not None:
            self.send_to_all(encoded)
        else:
            LOGGER.warning("Encoder of type %s encoded 'None' type object", type(self))
        return encoded

    @abc.abstractmethod
    def encode_api(self, data):
        """
        Encodes the given data and returns the result.

        This function allows for non-registered code to utilize the same
        serialization functionality as is used to encode data passed to the
        data_callback function.

        :param data: data to be encoded as rae bytes
        :return: encoded data bytes
        """



"""
Created on Apr. 27, 2015

@author: reder
"""

from fprime.common.models.serialize.type_base import BaseType

# Import the types this way so they do not need prefixing for execution.
from fprime.common.models.serialize.type_exceptions import (
    TypeException,
    TypeMismatchException,
)


class Channel:
    """
    Channel class is for deserialize channel telemetry value.
    Really this is a container since the type will have it's own deserialize
    implementation.
    """

    def __init__(
        self,
        name,
        channel_id,
        comp_name,
        ch_description,
        ch_type,
        ch_format_string,
        low_red,
        low_orange,
        low_yellow,
        high_yellow,
        high_orange,
        high_red,
    ):
        """
        Constructor
        """
        #
        # Make sure correct types are passed
        #
        if not isinstance(name, str):
            raise TypeMismatchException(str, type(name))

        if not isinstance(channel_id, int):
            raise TypeMismatchException(int, type(channel_id))

        if not isinstance(ch_description, str):
            raise TypeMismatchException(str, type(ch_description))

        if not isinstance(ch_type, BaseType):
            raise TypeMismatchException(BaseType, type(ch_type))

        # Initialize event internal variables
        self.__name = name
        self.__comp_name = comp_name
        self.__id = channel_id
        self.__ch_desc = ch_description
        self.__ch_type = ch_type
        self.__format_string = ch_format_string
        self.__low_red = low_red
        self.__low_orange = low_orange
        self.__low_yellow = low_yellow
        self.__high_yellow = high_yellow
        self.__high_orange = high_orange
        self.__high_red = high_red
        #
        self.__time_base = None
        self.__time_context = None
        self.__time_sec = None
        self.__time_usec = None
        #
        self.__changed = False

    def deserialize(self, ser_data, offset):
        """
        Deserialize event arguments
        :param ser_data: Binary input of the channel value.
        :param offset: offset in data to deserialize from
        :return: value from deserialized channel
        """
        try:
            self.__ch_type.deserialize(ser_data, offset)
            val = self.__ch_type.val
        except TypeException as e:
            print(f"Channel deserialize exception {e.getMsg()}")
            val = "ERR"

        #
        return val

    def setTime(self, time_base, time_context, time_sec, time_usec):
        """
        Channel telemetry time updater.
        """
        self.__time_base = time_base
        self.__time_context = time_context
        self.__time_sec = time_sec
        self.__time_usec = time_usec

    def getTime(self):
        """
        Return time tuple for UI updater use.
        """
        return (
            self.__time_base,
            self.__time_context,
            self.__time_sec,
            self.__time_usec,
        )

    def getName(self):
        return self.__name

    def getCompName(self):
        return self.__comp_name

    def getId(self):
        return self.__id

    def getChDesc(self):
        return self.__ch_desc

    def getType(self):
        return self.__ch_type

    def getTimeBase(self):
        return self.__time_base

    def getTimeContext(self):
        return self.__time_context

    def getTimeSec(self):
        return self.__time_sec

    def getTimeUsec(self):
        return self.__time_usec

    def getFormatString(self):
        return self.__format_string

    def getLowRed(self):
        return self.__low_red

    def getLowOrange(self):
        return self.__low_orange

    def getLowYellow(self):
        return self.__low_yellow

    def getHighYellow(self):
        return self.__high_yellow

    def getHighOrange(self):
        return self.__high_orange

    def getHighRed(self):
        return self.__high_red

    @property
    def changed(self):
        """
        change is True if recently updated.
        change is False if not changed.
        """
        return self.__changed

    @changed.setter
    def changed(self, ch):
        if not ch == False or not ch == True:
            ch = True
        self.__changed = ch



"""
@brief Class to store a specific channel telemetry reading

@date Created July 2, 2018
@author R. Joseph Paetz

@bug No known bugs
"""

from fprime.common.models.serialize import time_type
from fprime_gds.common.data_types import sys_data
from fprime_gds.common.utils.string_util import format_string_template


class ChData(sys_data.SysData):
    """
    The ChData class stores a specific channel telemetry reading.
    """

    def __init__(self, ch_val_obj, ch_time, ch_temp):
        """
        Constructor.

        Args:
            ch_val_obj: The channel's value at the given time. Should be an
                        instance of a class derived from the BaseType class
                        (with a deserialized data value) or None
            ch_temp: Channel template instance for this channel
            ch_time: Time the reading was made

        Returns:
            An initialized ChData object
        """
        super().__init__()
        self.id = ch_temp.get_id()
        self.val_obj = ch_val_obj
        self.time = ch_time
        self.template = ch_temp
        self.pkt = None

    @staticmethod
    def get_empty_obj(ch_temp):
        """
        Obtains a channel object that is empty (has a value of None)

        Args:
            ch_temp: (ChTemplate object) Template describing the channel

        Returns:
            A ChData Object with ch value of None
        """
        return ChData(None, time_type.TimeType(), ch_temp)

    def set_pkt(self, pkt):
        """
        Set the packet object to which this channel belongs (can be None)

        Args:
            pkt: The packet object to which these channels were transmitted in
        """
        self.pkt = pkt

    def get_pkt(self):
        """
        Return the packet object to which this channel belongs (could be None)

        Returns:
            The channel's packet
        """
        return self.pkt

    def get_val(self):
        """
        Return the channel value

        Returns:
            The channel reading
        """
        return None if self.val_obj is None else self.val_obj.val

    def get_val_obj(self):
        """
        Return the channel's value object

        Returns:
            The channel's value object containing the value (obj of a type
            inherited from TypeBase
        """
        return self.val_obj

    @staticmethod
    def get_csv_header(verbose=False):
        """
        Get the header for a csv file containing channel data

        Args:
            verbose: (boolean, default=False) Indicates if header should be for
                                              regular or verbose output

        Returns:
            Header for a csv file containing channel data
        """
        return "Time,Raw Time,Name,ID,Value\n" if verbose else "Time,Name,Value\n"

    def get_str(self, time_zone=None, verbose=False, csv=False):
        """
        Convert the channel data to a string

        Args:
            time_zone: (tzinfo, default=None) Timezone to print time in. If
                      time_zone=None, use local time.
            verbose: (boolean, default=False) Prints extra fields if True
            csv: (boolean, default=False) Prints each field with commas between
                                          if true

        Returns:
            String version of the channel data
        """
        time_str_nice = self.time.to_readable(time_zone)
        raw_time_str = str(self.time)
        ch_name = self.template.get_full_name()
        fmt_str = self.template.get_format_str()
        if self.val_obj is None:
            ch_val = "EMPTY CH OBJ"
        elif fmt_str:
            ch_val = format_string_template(fmt_str, (self.val_obj.val,))
        else:
            ch_val = str(self.val_obj.val)

        if verbose and csv:
            return "%s,%s,%s,%d,%s" % (
                time_str_nice,
                raw_time_str,
                ch_name,
                self.id,
                ch_val,
            )
        elif verbose and not csv:
            return "%s: %s (%d) %s %s" % (
                time_str_nice,
                ch_name,
                self.id,
                raw_time_str,
                ch_val,
            )
        elif not verbose and csv:
            return f"{time_str_nice},{ch_name},{ch_val}"
        else:
            return f"{time_str_nice}: {ch_name} = {ch_val}"

    def get_val_str(self):

        """
        Convert the value to a string, using the format specifier if provided
        """
        fmt_str = self.template.get_format_str()
        if self.val_obj is None:
            return ""
        elif fmt_str:
            return fmt_str % (self.val_obj.val)
        else:
            return str(self.val_obj.val)

    def __str__(self):
        """
        Convert the ch data to a string

        Returns:
            String version of the channel data
        """
        return self.get_str()



""" Versioning information for the GDS

This file contains versioning and compatibility definitions for the GDS tool set.
"""

# Currently the GDS is backwards compatible with all dictionaries. This MUST be bumped with each framework release to
# ensure compatibility
MINIMUM_SUPPORTED_FRAMEWORK_VERSION = (0, 0, 0)
MAXIMUM_SUPPORTED_FRAMEWORK_VERSION = (3, 0, 0)



"""
A file containing utilities for interacting with the Integration Test API
"""

import types
from typing import Any, Dict, List

from fprime_gds.common.data_types.ch_data import ChData
from fprime_gds.common.data_types.event_data import EventData
from fprime_gds.common.data_types.sys_data import SysData
from fprime_gds.common.pipeline.standard import StandardPipeline
from fprime_gds.common.templates.data_template import DataTemplate
from fprime_gds.common.testing_fw import predicates
from fprime_gds.common.testing_fw.api import IntegrationTestAPI
from fprime_gds.common.utils.config_manager import ConfigManager


def initialize_test_api(
    app_dictionary_path: str,
    log_path: str = None,
    server_ip: str = "127.0.0.1",
    server_port: int = 50050,
) -> IntegrationTestAPI:
    """
    Initializes an Integration Test API instance for use.

    Note that these MUST be manually disconnected when you're done using them,
    by calling "pipeline.disconnect()" and "api.teardown()", respectively.

    :param app_dictionary_path: A string path from the current working directory
        to the "<project name>AppDictionary.xml" file for the project you're
        using the API with
    :param log_path: A string path to where files should be logged, or "None" if
        you only want console output
    :param server_ip: The IP of the GDS server you want to connect to
    :param server_port: The port for the Test API on the GDS server
    """
    pipeline = StandardPipeline()
    pipeline.setup(ConfigManager(), app_dictionary_path, "/tmp")
    pipeline.connect(server_ip, server_port)

    # instantiate Test API (log_path of "None" will disable Test API logging)
    api = IntegrationTestAPI(pipeline, log_path)

    return pipeline, api


def get_upcoming_event(
    test_api: IntegrationTestAPI,
    search_filter: predicates.predicate,
    start_time="NOW",
    timeout: float = 5,
) -> EventData:
    """
    Returns the next event matching the given search filter that occurs after
    this is called. Times out after the given amount if no matching new events
    are found.

    :param test_api: An API instance that will be called to find the next event
    :param search_filter: A predicate each found event is tested against; if
        the event doesn't test "True" against this, we ignore it and keep
        searching
    :param start_time: An optional index or predicate to specify the earliest
        event time to search for
    :param timeout: The maximum time (in seconds) to wait for an event

    :return: The first "EventData" found that passes the filter, or "None" if no
        such event is found within time
    """
    event_filter = predicates.satisfies_all(
        [search_filter, predicates.event_predicate()]
    )
    # Test API only takes integer timeouts
    return test_api.find_history_item(
        event_filter, test_api.get_event_test_history(), start_time, int(timeout)
    )


def get_upcoming_channel(
    test_api: IntegrationTestAPI,
    search_filter: predicates.predicate,
    start_time="NOW",
    timeout: float = 5,
) -> ChData:
    """
    Returns the next telemetry update matching the given search filter that
    occurs after this is called. Times out after the given amount if no
    matching new updates are found.

    :param test_api: An API instance that will be called to find the next update
    :param search_filter: A predicate each found update is tested against; if
        the item doesn't test "True" against this, we ignore it and keep
        searching
    :param start_time: An optional index or predicate to specify the earliest
        update time to search for
    :param timeout: The maximum time (in seconds) to wait for an update

    :return: The first "ChData" found that passes the filter, or "None" if no
        such update is found within time
    """
    channel_filter = predicates.satisfies_all(
        [search_filter, predicates.telemetry_predicate()]
    )
    # Test API only takes integer timeouts
    return test_api.find_history_item(
        channel_filter, test_api.get_telemetry_test_history(), start_time, int(timeout)
    )


def get_item_list(
    item_dictionary: Dict[Any, DataTemplate],
    search_filter: predicates.predicate,
    template_to_data: types.FunctionType,
) -> List[SysData]:
    """
    Returns an ID-sorted list containing all the possible item types that could
    occur on the running F' instance, and information about each one.

    :param item_dictionary: A dictionary full of DataTemplate objects containing
        the events you want to list
    :param search_filter: A Test API predicate used to filter out which items
        to include in the returned list
    :param template_to_data: A callback function accepting a single DataTemplate
        object and returning a SysData object

    :return: An ID sorted list with all the items from the given dictionary
        AFTER being filtered
    """
    # Dictionary has DataTemplates by default, so convert them to SysData so
    # filtering will work properly (several predicates assume SysData types)
    event_data_list = list(
        map(lambda x: template_to_data(x[1]), item_dictionary.items())
    )

    # Filter by using the given predicate on the event values
    event_data_list = list(filter(search_filter, event_data_list))
    event_data_list.sort(key=lambda x: x.get_id())
    return event_data_list



"""
Created on Jan 5, 2015

@author: tcanham
"""

import copy
from enum import Enum

from fprime.common.models.serialize.type_base import BaseType

# Import the types this way so they do not need prefixing for execution.
from fprime.common.models.serialize.type_exceptions import (
    ArgLengthMismatchException,
    ArgNotFoundException,
    TypeMismatchException,
)
from fprime.common.models.serialize.numerical_types import U32Type

Descriptor = Enum(value="Descriptor", names="ABSOLUTE RELATIVE")


class Command:
    """
    classdocs
    """

    def __init__(
        self,
        component,
        mnemonic,
        opcode,
        description,
        arguments,
        seconds=0,
        useconds=0,
        descriptor=Descriptor.RELATIVE,
    ):
        """
        Constructor
        """

        # Make sure correct types are passed

        if not isinstance(component, str):
            raise TypeMismatchException(str, type(component))

        if not isinstance(mnemonic, str):
            raise TypeMismatchException(str, type(mnemonic))

        if not isinstance(opcode, int):
            raise TypeMismatchException(int, type(opcode))

        if not isinstance(description, str):
            raise TypeMismatchException(str, type(description))

        if not isinstance(arguments, list):
            raise TypeMismatchException(list, type(arguments))

        for (argname, argdesc, argtype) in arguments:
            #
            if not isinstance(argname, str):
                raise TypeMismatchException(str, type(argname))
            #
            if not isinstance(argdesc, str):
                raise TypeMismatchException(str, type(argdesc))
            #
            if not isinstance(argtype, BaseType):
                raise TypeMismatchException(BaseType, type(argtype))

        # Initialize command internal variables
        self.__component = component
        self.__mnemonic = mnemonic
        self.__opcode = opcode
        self.__description = description
        self.__arguments = arguments

        # If part of a sequence we need to set these as well
        self.setSeconds(seconds)
        self.setUseconds(useconds)
        self.setDescriptor(descriptor)

    def serialize(self):
        """
        Serializes command arguments
        """
        # first, serialize opcode
        opcode = U32Type(self.__opcode)
        ser_data = opcode.serialize()

        # then, serialize arguments
        for (arg_name, arg_desc, arg_type) in self.__arguments:
            ser_data += arg_type.serialize()
        return ser_data

    def getComponent(self):
        return self.__component

    def getMnemonic(self):
        return self.__mnemonic

    def getOpCode(self):
        return self.__opcode

    def getDescription(self):
        return self.__description

    def getSeconds(self):
        return self.__secs

    def getUseconds(self):
        return self.__usecs

    def getDescriptor(self):
        return self.__desc

    def getArgs(self):
        return self.__arguments

    def setSeconds(self, seconds):
        if not isinstance(seconds, int):
            raise TypeMismatchException(int, type(seconds))
        self.__secs = seconds

    def setUseconds(self, useconds):
        if not isinstance(useconds, int):
            raise TypeMismatchException(int, type(useconds))
        self.__usecs = useconds

    def setDescriptor(self, descriptor):
        if not isinstance(descriptor, type(Descriptor.ABSOLUTE)):
            raise TypeMismatchException(type(Descriptor.ABSOLUTE), type(descriptor))
        self.__desc = descriptor

    # value needs to be the Python class, not the raw value
    def setArg(self, arg_name, arg_type):
        """
        Method to set up an argument value of any
        of the types specified in serialize package.
        @param arg_name: string name of the argument.
        @param arg_type: object type to store argument value in.
        """
        # double check argument types
        if not isinstance(arg_name, str):
            raise TypeMismatchException(str, type(arg_name))

        if not isinstance(arg_type, BaseType):
            raise TypeMismatchException(BaseType, type(arg_type))

        new_arg_list = []
        found = False
        # search for argument
        for (arg, arg_desc, arg_value) in self.__arguments:
            if arg_name == arg:
                arg_value = arg_type
                found = True
            new_arg_list.append((arg, arg_desc, arg_value))

        if not found:
            raise ArgNotFoundException(arg_name)
        self.__arguments = new_arg_list

    def setArgs(self, values):
        """
        Given a list of values for the arguments, set the command arguments in order.
        @param values: raw python values (floats, ints, and strings) that will be converted
        into arg_types and stored as the command arg
        """

        # Make sure that the correct number of arguments is given
        if len(values) != len(self.__arguments):
            raise ArgLengthMismatchException(len(self.__arguments), len(values))

        # Set the new arguments by converting each value to an type:
        new_arg_list = []
        for value, (arg_name, arg_desc, arg_value) in zip(values, self.__arguments):
            new_value = copy.deepcopy(arg_value)
            new_value.val = value
            new_arg_list.append((arg_name, arg_desc, new_value))
        self.__arguments = new_arg_list



"""
Created on Feb. 9, 2015

@author: reder
"""

import traceback
from enum import Enum

from fprime.common.models.serialize.type_base import BaseType

# Import the types this way so they do not need prefixing for execution.
from fprime.common.models.serialize.type_exceptions import (
    TypeException,
    TypeMismatchException,
)

Severity = Enum(
    "Severity", "COMMAND ACTIVITY_LO ACTIVITY_HI WARNING_LO WARNING_HI DIAGNOSTIC FATAL"
)


class Event:
    """
    Event class is for deserializing log event messages.
    THis is essentially the equivalent of EVR's in MSL, SMAP, etc.
    """

    def __init__(
        self, name, event_id, severity, format_string, event_description, arguments
    ):
        """
        Constructor
        """

        # Make sure correct types are passed

        if not isinstance(name, str):
            raise TypeMismatchException(str, type(name))

        if not isinstance(event_id, int):
            raise TypeMismatchException(int, type(event_id))

        if not isinstance(format_string, str):
            raise TypeMismatchException(str, type(format_string))

        if not isinstance(event_description, str):
            raise TypeMismatchException(str, type(event_description))

        if not isinstance(arguments, list):
            raise TypeMismatchException(list, type(arguments))

        for (argname, argdesc, argtype) in arguments:
            #
            if not isinstance(argname, str):
                raise TypeMismatchException(str, type(argname))
            #
            if not isinstance(argdesc, str):
                raise TypeMismatchException(str, type(argdesc))
            #
            if not isinstance(argtype, BaseType):
                raise TypeMismatchException(BaseType, type(argtype))

        # Initialize event internal variables
        self.__name = name
        self.__id = event_id
        self.__severity = severity
        self.__format_string = format_string
        self.__event_description = event_description
        self.__arguments = arguments

    def deserialize(self, ser_data):
        """
        Deserialize event arguments
        @param ser_data: Binary input data of id followed by args
        """
        vals = []
        #
        # Next parse the arguments.
        #
        args = list(self.__arguments)
        offset = 0
        for arg in args:
            arg_obj = arg[2]
            try:
                # print arg_obj
                arg_obj.deserialize(ser_data, offset)
                vals.append(arg_obj.val)
            except TypeException as e:
                print(f"Event deserialize exception {e.getMsg()}")
                traceback.print_exc()
                vals.append("ERR")
            offset = offset + arg_obj.getSize()

        vals = [0] + vals
        return vals

    def stringify(self, event_args_list):
        """
        Return a formatted string of event args.
        """
        # print event_args_list
        return self.__format_string % tuple(event_args_list[1:])

    def getName(self):
        return self.__name

    def getId(self):
        return self.__id

    def getSeverity(self):
        return self.__severity

    def getFormatString(self):
        return self.__format_string

    def getEventDescription(self):
        return self.__event_description

    def getArgs(self):
        return self.__arguments



"""
flask/updown.py:

A simple service that handles file uploads and downloads. This allowed the REST api to show the status of file uplinks
and downlinks. In addition, an uplink destination directory is exposed for the UI to set where new uploads should be
uplinked to.

@author mstarch
"""
import os

import flask
import flask_restful


class Destination(flask_restful.Resource):
    """
    A data model for the current location of the destination of uplinked files.
    """

    def __init__(self, uplinker):
        """
        Constructor: setup the parser for incoming destination arguments
        """
        self.parser = flask_restful.reqparse.RequestParser()
        self.parser.add_argument(
            "destination", required=True, help="Destination to place uploaded files"
        )
        self.uplinker = uplinker

    def get(self):
        """
        Gets the current destination

        :return: current destination
        """
        return {"destination": self.uplinker.destination_dir}

    def put(self):
        """
        Updates the current destination directory
        """
        args = self.parser.parse_args()
        destination = args.get("destination", None)
        self.uplinker.destination_dir = destination
        return {"message": "success"}


class FileUploads(flask_restful.Resource):
    """
    A data model for the current uplinking file set.
    """

    def __init__(self, uplinker, uplink_set):
        """
        Constructor: setup the uplinker and argument parsing
        """
        self.uplinker = uplinker
        self.uplink_set = uplink_set
        self.parser = flask_restful.reqparse.RequestParser()
        self.parser.add_argument(
            "action", required=True, help="Action to take against files"
        )
        self.parser.add_argument(
            "source", required=False, default=None, help="File on which to act file"
        )

    def get(self):
        """
        Gets the current set of files

        :return: current uplinking files
        """
        return {
            "files": self.uplinker.current_files(),
            "running": self.uplinker.is_running(),
        }

    def put(self):
        """
        Handles an update to an existing source file.  Source and action are expected parameters to be supplied. If
        source is None, then "pause-all" or "unpause-all" should be supplied to globally pause the uplinker.
        """
        args = self.parser.parse_args()
        action = args.get("action", None)
        source = args.get("source", None)
        if action == "Remove" or action == "Cancel" and source is not None:
            self.uplinker.cancel_remove(source)
        elif action == "pause-all":
            self.uplinker.pause()
        elif action == "unpause-all":
            self.uplinker.unpause()

    def post(self):
        """
        Adds file(s) to be uplinked by enqueuing each into the uplinker.
        """
        successful = []
        failed = []
        for key, file in flask.request.files.items():
            try:
                filename = self.uplink_set.save(file)
                flask.current_app.logger.info(f"Received file. Saved to: {filename}")
                self.uplinker.enqueue(
                    os.path.join(self.uplink_set.config.destination, filename)
                )
                successful.append(key)
            except Exception as exc:
                flask.current_app.logger.warning(
                    f"Failed to save file {key} with error: {exc}"
                )
                failed.append(key)
        return {"successful": successful, "failed": failed}


class FileDownload(flask_restful.Resource):
    """ """

    def __init__(self, downlinker):
        """
        Constructor: setup the downlinker
        """
        self.downlinker = downlinker

    def get(self, source=None):
        """
        Gets the current downlinking files

        :return: current downlinking files
        """
        # Serve the  source if asked for, otherwise list all files
        if source is not None:
            return flask.send_from_directory(
                self.downlinker.directory, os.path.basename(source), as_attachment=True
            )
        else:
            return {"files": self.downlinker.current_files()}
